# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 17:52:28 2020

@author: willi
"""
import numpy as np
from . import CodonDictionaries
from . import translation_models as models
from . import ODE_Soln
from . import SSA_Soln
SSA_Soln = SSA_Soln.SSA_Soln
ODE_Soln = ODE_Soln.ODE_Soln

import time

import os

path_to_cpp = ''
path_to_gen = ''
path_to_trna = ''

for root, dirs, files in os.walk(".", topdown=False):
   for branch in dirs:
       if 'ssa_cpp' in branch:
           path_to_cpp = os.path.join(root, branch)
       if 'generalized_cpp' in branch:
           path_to_gen = os.path.join(root, branch)
       if 'trna_ssa' in branch:
           path_to_trna = os.path.join(root, branch)
if path_to_cpp != '':
    try:
        cwd = os.getcwd()
        os.chdir(path_to_cpp)
        
        import ssa_translation
        import ssa_translation_lowmem
        os.chdir(cwd)
    except:
        os.chdir(cwd)
    

if path_to_gen != '':
    try:
        cwd = os.getcwd()
        
        os.chdir(path_to_gen)
        print('importing C++ models')
        import ssa_translation_generic
        import ssa_translation_generic_lowmem
        print('c++ models loaded successfully')
        os.chdir(cwd)
    except:
        os.chdir(cwd)

if path_to_trna !='':
    try:
        cwd = os.getcwd()
        
        os.chdir(path_to_gen)
        print('importing C++ tRNA models')
        import ssa_trna
        print('c++ models loaded successfully')
        os.chdir(cwd)
    except:
        os.chdir(cwd)   
    

class TranslationSolvers():
    '''
    Container class for the solvers
    '''
    def __init__(self, time=None,xi=None):
        self.k = None
        self.k_bind = None
        self.k_term = None
        self.multiframe = False
        self.additional_rxns = {}
        self.probe_locations = None
        self.colors = 1
        
        try:
            ssa_translation_lowmem.np
            self.cython_available = True
        except:
            self.cython_available = False
        
        
        self.default_conditions = {'low_mem':True,
                                   'perturb':[0,0,0],
                                   'leaky_probes':False,
                                   'bins':None,
                                   'kprobe':1,
                                   'footprint':9,
                                   'burnin':0,
                                   'record_stats':False,
                                   'kon':1.,
                                   'koff':1.,
                                   'n_traj':30,
                                   'bursting':False,
                                   'tRNA':False,
                                   'ribosome_preallocation':100,
                                   'dynamic_rib_prealloc':False
                                   }
        
        self.t = np.linspace(0,1000,1001)
        self.x0 = []
        self._poi = None
    
    @property
    def protein(self):
        return self._poi

    @protein.setter         #Auto detect amount of colors if protein is set
    def protein(self,newpoi):
        self._poi = newpoi
        try:
            self.colors = self._poi.probe_loc.shape[0]
        except:
            pass
        

    def solve_ballistic_model(self, ki,ke, poi=None, tag= None):
        '''
        

        Parameters
        ----------
        ki : float
            initation rate of ribosomes.
        ke : float
            mean elongation rate along the transcript.
        poi : protein object, optional
            a particular protein object to pull geometry from. The default is None.
        tag : list, optional
            list of tag locations if geometry not given. The default is None.

        Returns
        -------
        tau_analyticals : list of floats
            analytically sovled decorrelation times tau.
        mean_analyticals : list of floats
            the visible ribosome means on the transcript.
        var_analyticals : list of floats
            variances of ribosomes on the transcript.

        '''
   
        if poi == None:
            poi = self._poi
        if tag == None:
            colors = np.where((poi.probe_loc)== 1)[0]
            locs = np.where((poi.probe_loc)== 1)[1]
            tags = []
            for i in range(0,max(colors)+1):
                
                tags.append(locs[np.where(colors == i)].tolist())
            
            
        tau_analyticals = []
        mean_analyticals = []
        var_analyticals = []
        for tag in tags:

            
            L = poi.total_length #get the total length of the gene
            Lm = np.mean(tag)  #the mean location of the tag epitopes
            L_after_tag = L - tag[-1]
            L_tag = int((tag[-1] - tag[0]) / 2)

            ke_analytical = L*ke / np.sum(self.__get_ui(poi.nt_seq[:-3]))

            tau_analytical = (L )/ke_analytical  #analytical tau ie autocovariance time 
            mean_analytical = ki*tau_analytical * (1.-Lm/float(L)) # mean intensity
            var_analytical = ki*tau_analytical * (1.-Lm/float(L))**2  #var intensity
            
            tau_analyticals.append(tau_analytical)
            mean_analyticals.append(mean_analytical)
            var_analyticals.append(var_analytical)
        
        return tau_analyticals,mean_analyticals,var_analyticals
    
    
    def invert_ballistic(self,tau_measured, mu_I, poi= None):
        '''

        Parameters
        ----------
        tau_measured : float, int, list, ndarray
            measured analytical decorrelation time(s).
        mu_I : float, int, list, ndarray
            average intensitie(s).
        poi : Poi object (optional)
            Protein of interest to pull geometry from.

        Returns
        -------
        kes : list
            analytical average k_elongation(s).
        kis : list
            analytical average k_initations(s).

        '''
        if poi == None:
            poi = self._poi
        #if tag == None:
        colors = np.where((poi.probe_loc)== 1)[0]
        locs = np.where((poi.probe_loc)== 1)[1]
        
        tags = []
        
        if isinstance(tau_measured,np.ndarray):
            tau_measured = tau_measured.tolist()
        
        if not isinstance(tau_measured,list):
            tau_measured = [tau_measured]
            
            
        if isinstance(mu_I,np.ndarray):
            mu_I = mu_I.tolist()
        
        if not isinstance(mu_I,list):
            mu_I = [mu_I]
            
            
        for i in range(0,max(colors)+1):
            
            tags.append(locs[np.where(colors == i)].tolist())
            
        kes = []
        kis = []
        
        for i in range(len(tags)):
            tag= tags[i]
            print(tag)
            L = poi.total_length #get the total length of the gene
            Lm = np.mean(tag)  #the mean location of the tag epitopes
            L_after_tag = L - tag[-1]
            L_tag = int((tag[-1] - tag[0]) / 2)
        
            ke_analytical = (L)/tau_measured[i]
            
            ke = ke_analytical/(L)*np.sum(self.__get_ui(poi.nt_seq[:-3]))
            
            kes.append(ke)
            ki = (mu_I[i]/len(tag)) / ( (1.-Lm/float(L))*tau_measured[i])
            kis.append(ki)
            
        return kes, kis


    
    def __get_ui(self, nt_seq):
        '''
        return the ratio of average gene copy number / sequence codon copy number
        '''
        
        codon_dict = CodonDictionaries()
        mean_u = np.mean(list(codon_dict.strGeneCopy_single.values()) )
        ui = []
        for i in range(0, len(nt_seq), 3):
            ui.append(mean_u/ codon_dict.strGeneCopy[nt_seq[i:i+3]])
        return ui
        

    def solve_ssa_set_conditions(self):
        
        ssa_conditions = self.default_conditions
        kprobe = self.default_conditions['kprobe']
        kon = self.default_conditions['kon']
        koff = self.default_conditions['koff']
        
        if kprobe != 1:
            ssa_conditions['leaky_probes'] = True
            leaky_probes = True
        
        perturb = ssa_conditions['perturb']
        leaky_probes = ssa_conditions['leaky_probes'] 
        low_memory = ssa_conditions['low_mem'] 
        record_stats = ssa_conditions['record_stats'] 
        bins = ssa_conditions['bins']
        n_traj = ssa_conditions['n_traj']
        
        
        probe_vec = self.protein.probe_vec.astype(np.int32)
        t = self.t
        
        k = [self.protein.ki,] + self.protein.kelong + [self.protein.kt,]
        self.__check_rates(k)
        x0 = self.x0
        
        
        if self.cython_available:
            if low_memory:
                ssa_obj = self.__solve_ssa_lowmem_combined(k,t,x0,n_traj,ssa_conditions = ssa_conditions, kon=kon, koff=koff, kprobe=kprobe)
            else:
            
                if record_stats:
                
                    if leaky_probes == False:
                        # if low_memory:
                        #     ssa_obj = self.__solve_ssa_lowmem(k,t,x0,n_traj, ssa_conditions = ssa_conditions)
                        # else:
                        ssa_obj = self.__solve_ssa(k,t,x0,n_traj,ssa_conditions = ssa_conditions)
                    else:
                        # if low_memory:
                        #     ssa_obj = self.__solve_ssa_lowmem_leaky(k,t,x0,k_probe,n_traj, ssa_conditions = ssa_conditions)
                        # else:
                        ssa_obj = self.__solve_ssa_leaky(k,t,x0,kprobe,n_traj,ssa_conditions = ssa_conditions)         
                   
                else:
                    if leaky_probes == False:
                        # if low_memory:
                        #     ssa_obj = self.__solve_ssa_lowmem_nostats(k,t,x0,n_traj, ssa_conditions = ssa_conditions)
                        # else:
                        ssa_obj = self.__solve_ssa_nostats(k,t,x0,n_traj,ssa_conditions = ssa_conditions)
                    else:
                        # if low_memory:
                        #     ssa_obj = self.__solve_ssa_lowmem_leaky_nostats(k,t,x0,k_probe,n_traj, ssa_conditions = ssa_conditions)
                        # else:
                        ssa_obj = self.__solve_ssa_leaky_nostats(k,t,x0,kprobe,n_traj,ssa_conditions = ssa_conditions)         
                           
        else:
            ssa_obj = self.__solve_ssa_python(k,t,x0,n_traj,ssa_conditions = ssa_conditions, kon=kon, koff=koff, kprobe=kprobe)


        
       
        return ssa_obj
            

    def solve_ssa_trna(self,k_index, k_diffusion, k_bind, kelong, k_compl, t,x0=[], k_trna = None, perturb=[0,0,0],leaky_probes=False,kprobe=np.ones(1),probe_vec = None, probe_loc=None, kon=1,koff=1,bursting=False,n_traj=10   ):
        self.__check_rates_trna(k_index)
        
        ssa_conditions = self.default_conditions
        if np.sum(kprobe != 1) !=0:
            ssa_conditions['leaky_probes'] = True
            leaky_probes = True
        
        ssa_conditions['perturb'] = perturb
        ssa_conditions['leaky_probes'] = leaky_probes

        
        ssa_conditions['bursting'] = bursting
        
        provided_probe = False
        try:
            probe_vec[0]
            provided_probe = True
        except:
            pass
        
        provided_protein = False
        try:
            self.protein.kelong[0]
            provided_protein = True
        except:
            pass       
            
        try:
            k_trna[0]
        except:
            
            strGeneCopy = CodonDictionaries().strGeneCopy_single
            strGeneCopy.pop('TAG')
            strGeneCopy.pop('TAA')
            strGeneCopy.pop('TGA')

            k_trna = np.array(list(CodonDictionaries().strGeneCopy_single.values()))
        
        print(k_trna)
        if not provided_probe:
            if provided_protein:
                probe_vec = self.protein.probe_vec.astype(np.int32)
                probe_loc = self.protein.probe_loc.astype(np.int32)
            else:
                print("no provided probe vector, please set the solver.protein with a protein object or provide a probe vector")
                raise 
        else:
            probe_vec = probe_vec
            
            
        ssa_conditions['probe_vec'] = probe_vec
        ssa_conditions['probe_loc'] = probe_loc
        
        
     
        ssa_obj = self.__solve_ssa_trna(k_index,k_trna, k_diffusion,k_bind,kelong, k_compl,t,x0,n_traj,ssa_conditions = ssa_conditions)
                        
        return ssa_obj
    
    def solve_ssa(self,ke,t,ki=.33,kt = 10,x0=[],n_traj=100,bins=None,low_memory=True,perturb=[0,0,0],leaky_probes=False,kprobe=np.ones(1),record_stats=False,probe_vec = None, probe_loc=None, kon=1,koff=1,bursting=False, rib_prealloc=200, dynamic_prealloc=False):
        
        self.__check_rates(ke)
        
        ssa_conditions = self.default_conditions
        if np.sum(kprobe != 1) !=0:
            ssa_conditions['leaky_probes'] = True
            leaky_probes = True
        
        ssa_conditions['perturb'] = perturb
        ssa_conditions['leaky_probes'] = leaky_probes
        ssa_conditions['low_mem'] = low_memory
        ssa_conditions['record_stats'] = record_stats
        ssa_conditions['bins'] = bins
        ssa_conditions['bursting'] = bursting
        ssa_conditions['ribosome_preallocation'] = rib_prealloc
        ssa_conditions['dynamic_rib_prealloc'] = dynamic_prealloc

        provided_probe = False
        
        try:
            probe_vec[0]
            provided_probe = True
        except:
            pass
        
        provided_protein = False
        try:
            self.protein.kelong[0]
            provided_protein = True
            
        except:
            pass       
        
        if provided_protein: #parse out stop codons if they are included
            includes_stop = False
            if self.protein.aa_seq[-1] == '*' or self.protein.nt_seq[-1].upper() in ['UAA','TAA','TAG','UAG', 'UGA','TGA']:
                if len(self.protein.aa_seq) == len(ke):
                    includes_stop = True

            

        if not provided_probe:
            if provided_protein:
                if includes_stop:
                    probe_vec = np.require(self.protein.probe_vec.astype(np.int32)[:,:-1],requirements=['C'])
                    probe_loc = np.require(self.protein.probe_loc.astype(np.int32)[:,:-1],requirements=['C'])
                    ke = ke[:-1]
                else:
                    probe_vec = self.protein.probe_vec.astype(np.int32)
                    probe_loc = self.protein.probe_loc.astype(np.int32)
            else:
                print("no provided probe vector, please set the solver.protein with a protein object or provide a probe vector")
                raise 
        else:
            probe_vec = probe_vec
            
            
        ssa_conditions['probe_vec'] = probe_vec
        ssa_conditions['probe_loc'] = probe_loc
   
        
        if dynamic_prealloc:
            
            ssa_conditions['ribosome_preallocation'] = int(len(ke)/ssa_conditions['footprint'])+5
   
        k = [ki,] + ke + [kt,]
        if self.cython_available:
            if low_memory:
                ssa_obj = self.__solve_ssa_lowmem_combined(k,t,x0,n_traj,ssa_conditions = ssa_conditions, kon=kon, koff=koff, kprobe=kprobe)
            else:
                ssa_obj = self.__solve_ssa_full_combined(k,t,x0,n_traj,ssa_conditions = ssa_conditions, kon=kon, koff=koff, kprobe=kprobe)
 
            
        else:
            ssa_obj = self.__solve_ssa_python(k,t,x0,n_traj,ssa_conditions = ssa_conditions, kon=kon, koff=koff, kprobe=kprobe)

       
       #Generate metadata 
        colors = ssa_obj.intensity_vec.shape[0]
        st = ssa_obj.start_time
        ft = ssa_obj.time_rec[-1]
        lt = len(ssa_obj.time_rec)
        if bins == None:
            bstr = 0
        else:
            bstr=1
        sid = 'L' + str(len(k)) + 'N' + str(n_traj) + 'T' + str(st) + '_' + str(ft) + '_' + str(lt) 
        fstr = '-F' + str(int(self.cython_available)) + str(int(sum(perturb))) + str(int(leaky_probes)) + str(int(low_memory))  + str(int(record_stats))  +  str(int(bstr)) + str(int(bursting)) 
        cstr = 'C' + str(int(colors))
        nprobe = np.sum(probe_loc,axis=1)
        for value in nprobe:
            cstr = cstr+ 'P' + str(int(value))
        sid = sid + cstr + fstr
        ssa_obj._SSA_Soln__meta['id'] = sid
        
        return ssa_obj
    
    

    def solve_ode(self,k,t,x0,kbind,pl,bins=None, corr = False):
        st = time.time()
        m_ode = models.TranslateODE()
        m_ode.N = len(k)
        m_ode.tf = t[-1]
        m_ode.ptimes = len(t)
        m_ode.ke = k
        m_ode.kb = kbind 
        m_ode.fi = 1
        m_ode.ti = t[0]
        m_ode.binary = pl
        
        #mu_intensity = m_ode.solve()
        
        m = models.TranslateCorrs()
        m.N = len(k)
        m.tf = t[-1]
        m.ptimes = len(t)
        m.ke = k
        m.kb = kbind 
        m.fi = 1
        m.ti = t[0]
        m.xi = x0
        m.binary = pl
        
        
        ode_soln = ODE_Soln()
              
        m.get_autonomous_matrix()
        #m.faster_solve()
       
        m.get_mean_SS()
        

        
        mean_ss = m.get_mean_SS()
        
        mean_I = m.map_to_fluorescence3(mean_ss)
        m.mu_ss = mean_ss
    
        var_ss = m.get_var_SS()
        var_I = m.map_to_fluorescence(var_ss) 
        
        
        if corr:
            
            m.csolve()
            norm_acc = np.ravel((m.intensity)/var_I)
            intensity_acc = m.intensity
            ode_soln.intensity_acc = m.intensity
            ode_soln.intensity_acc_norm = norm_acc
            

        ode_soln.mu_state_ss = mean_ss
        ode_soln.var_state_ss = var_ss

        ode_soln.mu_It = m.solve()
        ode_soln.mu_I_ss = mean_I
        ode_soln.var_I_ss = var_I
        ode_soln.time = t
        
        ode_soln.N = len(k)
        ode_soln.k = k
        ode_soln.x0 = x0
        ode_soln.fi = 1
        ode_soln.kb = kbind
        ode_soln.prob_loc = pl
        ode_soln.solve_time = time.time()-st
        
        if len(k) > 100:
            ode_soln.solve_method = 'expv'
        else:
            ode_soln.solve_method = 'odeint'
        
        
        return ode_soln
    
    
    
    
    def __solve_ssa(self,k,t,x0,n_traj,ssa_conditions=None):
        
        seeds = np.random.randint(0, 0x7FFFFFF, n_traj, dtype = np.int32)
        
        if ssa_conditions == None:
            ssa_conditions = self.default_conditions
        
        x0 = self.__check_x0(x0)
   
       
        rib_vec = []
        solutions = []            
        solutionssave = []       
        N_rib = 200
        colors = self.colors
        all_results,all_nribs,all_collisions,all_frapresults,all_ribtimes,all_col_points = self.__generate_mats(n_traj,k[0],t,N_rib,colors)
        footprint = ssa_conditions['footprint']
        evf = ssa_conditions['perturb'][0]
        evi = ssa_conditions['perturb'][1]
        intime = ssa_conditions['perturb'][2]
        non_consider_time = ssa_conditions['burnin']
        
        print(t)
        
        st = time.time()
        
        for i in range(n_traj):
            
            result,ribtimes,frapresult,coltimes,colpointsx,colpointst = self.__generate_vecs(k,t,N_rib,colors)
            nribs = np.array([0],dtype=np.int32)
            kelong = np.array(k[1:-1]).flatten()
            

    

            ssa_translation.run_SSA(result, ribtimes, coltimes, colpointsx,colpointst, kelong,frapresult, t, k[0], k[-1], evf, evi, intime, seeds[i],nribs,x0,footprint,200)
            #ssa_translation.run_SSA(result, ribtimes, coltimes, k[1:-1],frapresult, truetime, k[0], k[-1], evf, evi, intime, seeds[i],nribs)
            
            all_results[i, :] = result.T
            all_frapresults[i,:] = frapresult
            all_ribtimes[i,:] = ribtimes
            all_collisions[i,:] = coltimes
            all_nribs[i,:] = nribs
            
            endcolrec = np.where(colpointsx == 0)[0][0]
            
            colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
            all_col_points.append(colpoints.T)
                
            
                
        maxso = 0    
        for i in range(n_traj):
            soln = all_results[i, :].reshape((N_rib, len(t)))
       

            validind = np.where(np.sum(soln,axis=1)!=0)[0]
            

            if np.max(validind) != N_rib-1:
                validind = np.append(np.where(np.sum(soln,axis=1)!=0)[0],np.max(validind)+1)
                
            so = soln[(validind,)]
            if so.shape[0] > maxso:
                maxso = so.shape[0]
                
                
                
        for i in range(n_traj):
            soln = all_results[i, :].reshape((N_rib, len(t)))
       

            validind = tuple([x for x in range(0,maxso)])
            

            so = soln[(validind,)]
                
            solutionssave.append(so)
            solutions.append(soln)
            
        
        collisions = np.array([[]])
        watched_ribs = []
        for i in range(n_traj):
            totalrib = all_nribs[i]
        
            if totalrib > all_collisions.shape[1]:
                collisions = np.append(collisions, all_collisions[i][:])
                watched_ribs.append(int(all_collisions.shape[1]))
        
            else:
               
                collisions = np.append(collisions, all_collisions[i][:int(totalrib[0])])
                watched_ribs.append(int(totalrib[0]))
        
        sttime = time.time() - st


        no_ribosomes = np.zeros((n_traj, (len(k)+1)))
        
        startindex = np.where(t >= non_consider_time)[0][0]
        
        #all_results = all_results[:,startindex*N_rib:]
        pv = self.protein.probe_vec
        I = np.zeros((colors,len(t), n_traj))
        
        for n in range(colors):
            for i in range(n_traj):
                traj = all_results[i,:].reshape((N_rib,len(t))).T
                for j in range(len(t)):
                    temp_output = traj[j,:]
    
                    I[n,j,i] = np.sum(pv[n][temp_output[temp_output>0]-1]  )

        for i in range(len(solutions)):
            for j in range(len(solutions[0][0][startindex:])):
                rib_pos = solutions[i][startindex:, j][np.nonzero(solutions[i][startindex:, j])]
               
                no_ribosomes[i, rib_pos.astype(int)] += 1
        no_ribosomes = no_ribosomes[:, 1:]

        ribosome_means = np.mean(no_ribosomes, axis=0)
        ribosome_density = ribosome_means/len(k)

        no_ribosomes_per_mrna = np.mean(no_ribosomes)
        
        ssa_obj = SSA_Soln()
        ssa_obj.no_ribosomes = no_ribosomes
        ssa_obj.n_traj = n_traj
        ssa_obj.k = k
        ssa_obj.no_rib_per_mrna = no_ribosomes_per_mrna
        ssa_obj.rib_density = ribosome_density
        ssa_obj.rib_means = ribosome_means
        ssa_obj.rib_vec = rib_vec
        #ssa_obj.intensity_vec = intensity_vec
        ssa_obj.time_vec_fixed = t
        ssa_obj.time = t
        ssa_obj.time_rec = t[startindex:]
        ssa_obj.start_time = non_consider_time
        ssa_obj.watched_ribs = watched_ribs
        ssa_obj.intensity_vec = I
        ssa_obj.solutions = np.array(solutionssave)
        try:
            ssa_obj.col_points = all_col_points
        except:
            pass
        
        
        
        return ssa_obj
    

    
                
    def __map_to_intensity(self): 
        
        return 1           


    def __solve_ssa_trna(self,kindex,ktrna,kdiffusion,kbind,kelong,kcompl, t, x0, n_traj, ssa_conditions=None ):

        

        seeds = np.random.randint(0, 0x7FFFFFF, n_traj, dtype = np.int32)
    
        if ssa_conditions == None:
            ssa_conditions = self.default_conditions
        
        x0 = self.__check_x0(x0)
   
       
        rib_vec = []
        solutions = []            
        solutionssave = []       
        N_rib = 200
        colors = self.colors
        
        n_trajectories = n_traj
        

        all_results,all_nribs,all_collisions,all_frapresults,all_ribtimes,all_col_points = self.__generate_mats(n_traj,kbind,t,N_rib,colors)
        all_trna_results = np.zeros((n_trajectories,61*len(t)),dtype=np.int32)
        
        footprint = ssa_conditions['footprint']
        evf = ssa_conditions['perturb'][0]
        evi = ssa_conditions['perturb'][1]
        intime = ssa_conditions['perturb'][2]
        non_consider_time = ssa_conditions['burnin']
        
        st = time.time()
        
        N_rib = 200
        
       
        result = np.zeros((len(t)*N_rib),dtype=np.int32  )
        
        #lenfrap = len(np.intersect1d(np.where(t_array>0)[0],np.where(t<20)[0]))
        #all_frapresults = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)

        all_ribs = np.zeros((n_trajectories,1))
        all_col_points = []
    
        #seeds = np.random.randint(0,0x7FFFFFF,n_trajectories)
        x0 = np.zeros((N_rib),dtype=np.int32)
     
        
        for i in range(n_trajectories):
            
            trna_result = np.zeros((len(t)*61),dtype=np.int32)    
            result,ribtimes,frapresult,coltimes,colpointsx,colpointst = self.__generate_vecs_trna(kbind,kindex,t,N_rib,colors)

            nribs = np.array([0],dtype=np.int32)
            
            ssa_trna.run_SSA(result,trna_result,ribtimes,coltimes,colpointsx,colpointst, kindex,ktrna,kdiffusion,frapresult,t,kbind,kcompl, 0,0,0, seeds[i],nribs,x0,kelong)
                  
            all_results[i, :] = result.T
            all_trna_results[i,:] = trna_result
            all_frapresults[i,:] = frapresult
            all_ribtimes[i,:] = ribtimes
            all_collisions[i,:] = coltimes
            all_nribs[i,:] = nribs[0]
         
            endcolrec = np.where(colpointsx == 0)[0][0]
            
            colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
            all_col_points.append(colpoints.T)
                
        maxso = 0    
        for i in range(n_traj):
            soln = all_results[i, :].reshape((N_rib, len(t)))
       

            validind = np.where(np.sum(soln,axis=1)!=0)[0]
            

            if np.max(validind) != N_rib-1:
                validind = np.append(np.where(np.sum(soln,axis=1)!=0)[0],np.max(validind)+1)
                
            so = soln[(validind,)]
            if so.shape[0] > maxso:
                maxso = so.shape[0]
                
                
                
        for i in range(n_traj):
            soln = all_results[i, :].reshape((N_rib, len(t)))
       

            validind = tuple([x for x in range(0,maxso)])
            

            so = soln[(validind,)]
                
            solutionssave.append(so)
            solutions.append(soln)
            
            
        
        
        collisions = np.array([[]])
        watched_ribs = []
        for i in range(n_traj):
            totalrib = all_nribs[i]
        
            if totalrib > all_collisions.shape[1]:
                collisions = np.append(collisions, all_collisions[i][:])
                watched_ribs.append(int(all_collisions.shape[1]))
        
            else:
               
                collisions = np.append(collisions, all_collisions[i][:int(totalrib[0])])
                watched_ribs.append(int(totalrib[0]))
        
        sttime = time.time() - st


        no_ribosomes = np.zeros((n_traj, (len(kindex)+1)))
        
        startindex = np.where(t >= non_consider_time)[0][0]
        
        #all_results = all_results[:,startindex*N_rib:]
        pv = self.protein.probe_vec
        I = np.zeros((colors,len(t), n_traj))
        
        for n in range(colors):
            for i in range(n_traj):
                traj = all_results[i,:].reshape((N_rib,len(t))).T
                for j in range(len(t)):
                    temp_output = traj[j,:]
    
                    I[n,j,i] = np.sum(pv[n][temp_output[temp_output>0]-1]  )

        for i in range(len(solutions)):
            for j in range(len(solutions[0][0][startindex:])):
                rib_pos = solutions[i][startindex:, j][np.nonzero(solutions[i][startindex:, j])]
               
                no_ribosomes[i, rib_pos.astype(int)] += 1
        no_ribosomes = no_ribosomes[:, 1:]

        ribosome_means = np.mean(no_ribosomes, axis=0)
        ribosome_density = ribosome_means/len(kindex)

        no_ribosomes_per_mrna = np.mean(no_ribosomes)
        
        ssa_obj = SSA_Soln()
        ssa_obj.no_ribosomes = no_ribosomes
        ssa_obj.n_traj = n_traj
        ssa_obj.k = kindex
        ssa_obj.no_rib_per_mrna = no_ribosomes_per_mrna
        ssa_obj.rib_density = ribosome_density
        ssa_obj.rib_means = ribosome_means
        ssa_obj.rib_vec = rib_vec
        ssa_obj.I = I
        ssa_obj.eval_time = sttime
        ssa_obj.time_vec_fixed = t
        ssa_obj.time = t
        ssa_obj.time_rec = t[startindex:]
        ssa_obj.start_time = non_consider_time
        ssa_obj.watched_ribs = watched_ribs
        ssa_obj.intensity_vec = I
        ssa_obj.solutions = np.array(solutionssave)
        ssa_obj.all_trna_results = all_trna_results
        try:
            ssa_obj.col_points = all_col_points
        except:
            pass
        
        
        
        return ssa_obj                

    def __solve_ssa_python(self,k,t,x0,n_traj,ssa_conditions=None, kprobe=None, kon=None, koff=None, flags=[0,0,0]):
        
        seeds = np.random.randint(0, 0x7FFFFFF, n_traj, dtype=np.int32)
        
        if ssa_conditions == None:
            ssa_conditions = self.default_conditions
        
        x0 = self.__check_x0(x0)
   
        flags = [int(ssa_conditions['bursting']), int(ssa_conditions['leaky_probes']), int(ssa_conditions['record_stats'])]

        rib_vec = []
        solutions = []            
        solutionssave = []       
        N_rib = 200
        colors = self.colors
        
        all_results,all_nribs,all_collisions,all_frapresults,all_ribtimes,all_col_points = self.__generate_mats(n_traj,k[0],t,N_rib,colors)
        footprint = ssa_conditions['footprint']
        evf = ssa_conditions['perturb'][0]
        evi = ssa_conditions['perturb'][1]
        intime = ssa_conditions['perturb'][2]
        non_consider_time = ssa_conditions['burnin']
        
        
        st = time.time()
        
        
        print('C++ library failed, Using Python Implementation')
        rib_vec = []

        solutions = []            
        solutionssave = []
        N_rib = ssa_conditions['ribosome_preallocation']
        collisions = np.array([[]])
        all_results = np.zeros((n_traj, N_rib*len(t)), dtype=np.int32)
        I_internal = np.zeros((colors,len(t), n_traj)) 
        all_col_points = []
        watched_ribs = []
        for i in range(n_traj):
            
            soln,all_ribtimes,Ncol,col_points,intensity = self.__ssa_python(k, t, inhibit_time=intime+non_consider_time, FRAP=evf, Inhibitor=evi, flags=flags, kon=kon, kprobe=kprobe, koff=koff, ssa_conditions=ssa_conditions)
            #soln = soln.reshape((1, (len(time_vec_fixed)*N_rib)))
            
            collisions = np.append(collisions,Ncol)
            watched_ribs.append(int(len(collisions)))
            validind = np.where(np.sum(soln,axis=1)!=0)[0]
            all_col_points.append(np.array(col_points))
            if np.max(validind) != N_rib-1:
                validind = np.append(np.where(np.sum(soln,axis=1)!=0)[0],np.max(validind)+1)
            
            so = soln[(validind,)]
          
            #solutionssave.append(so)

            solutions.append(soln)
        
            result = soln.reshape((1, (len(t)*N_rib)))
            all_results[i, :] = result
            I_internal[:,:,i] = intensity
        
                        
        maxso = 0    
        for i in range(n_traj):
            soln = all_results[i, :].reshape((N_rib, len(t)))
      
            validind = np.where(np.sum(soln,axis=1)!=0)[0]
            if np.max(validind) != N_rib-1:
                validind = np.append(np.where(np.sum(soln,axis=1)!=0)[0],np.max(validind)+1)
                
            so = soln[(validind,)]
            if so.shape[0] > maxso:
                maxso = so.shape[0]
                     
        for i in range(n_traj):
            soln = all_results[i, :].reshape((N_rib, len(t)))
       
            validind = tuple([x for x in range(0,maxso)])
            so = soln[(validind,)]
                
            solutionssave.append(so)
            
            
        # for i in range(n_traj):
        #     # result,ribtimes,frapresult,coltimes,colpointsx,colpointst = self.__generate_vecs(k,t,N_rib,colors)
        #     # nribs = np.array([0],dtype=np.int32)
        #     # kelong = np.array(k[1:-1]).flatten()
            
        #     soln,all_ribtimes,Ncol,col_points  = self.__ssa_python(k,t,inhibit_time=intime+non_consider_time,FRAP=evf,Inhibitor=evi,flags=flags)
              
        #     all_results[i, :] = result.T
        #     all_frapresults[i,:] = frapresult
        #     all_ribtimes[i,:] = ribtimes
        #     all_collisions[i,:] = coltimes
        #     all_nribs[i,:] = nribs
            
        #     endcolrec = np.where(colpointsx == 0)[0][0]
            
        #     colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
        #     all_col_points.append(colpoints.T)
                
            
        # for i in range(n_traj):
        #     soln = all_results[i, :].reshape((N_rib, len(t)))
       

        #     validind = np.where(np.sum(soln,axis=1)!=0)[0]

        #     if np.max(validind) != N_rib-1:
        #         validind = np.append(np.where(np.sum(soln,axis=1)!=0)[0],np.max(validind)+1)
        
        #     so = soln[(validind,)]
            
        #     solutionssave.append(so)
        #     solutions.append(soln)
        
        collisions = np.array([[]])
        watched_ribs = []
        for i in range(n_traj):
            totalrib = all_nribs[i]
        
            if totalrib > all_collisions.shape[1]:
                collisions = np.append(collisions, all_collisions[i][:])
                watched_ribs.append(int(all_collisions.shape[1]))
        
            else:
               
                collisions = np.append(collisions, all_collisions[i][:int(totalrib[0])])
                watched_ribs.append(int(totalrib[0]))
        
        sttime = time.time() - st


        no_ribosomes = np.zeros((n_traj, (len(k)+1)))
        
        startindex = np.where(t >= non_consider_time)[0][0]
        
        #all_results = all_results[:,startindex*N_rib:]
        pv = self.protein.probe_vec
        # I = np.zeros((colors,len(t), n_traj))
        
        # for n in range(colors):
        #     for i in range(n_traj):
        #         traj = all_results[i,:].reshape((N_rib,len(t))).T
        #         for j in range(len(t)):
        #             temp_output = traj[j,:]
    
        #             I[n,j,i] = np.sum(pv[n][temp_output[temp_output>0]-1]  )
    
        for i in range(len(solutions)):
            for j in range(len(solutions[0][0][startindex:])):
                
                rib_pos = solutions[i][startindex:, j][np.nonzero(solutions[i][startindex:, j])]
                if len(rib_pos.astype(int)) > 0:
                    no_ribosomes[i, rib_pos.astype(int)] += 1
                
                    
        no_ribosomes = no_ribosomes[:, 1:]

        ribosome_means = np.mean(no_ribosomes, axis=0)
        ribosome_density = ribosome_means/len(k)

        no_ribosomes_per_mrna = np.mean(no_ribosomes)
        
        ssa_obj = SSA_Soln()
        ssa_obj.no_ribosomes = no_ribosomes
        ssa_obj.n_traj = n_traj
        ssa_obj.k = k
        ssa_obj.no_rib_per_mrna = no_ribosomes_per_mrna
        ssa_obj.rib_density = ribosome_density
        ssa_obj.rib_means = ribosome_means
        ssa_obj.rib_vec = rib_vec
        #ssa_obj.intensity_vec = intensity_vec
        ssa_obj.time_vec_fixed = t
        ssa_obj.time = t
        ssa_obj.time_rec = t[startindex:]
        ssa_obj.start_time = non_consider_time
        ssa_obj.watched_ribs = watched_ribs
        ssa_obj.intensity_vec = I_internal
        ssa_obj.I = I_internal
        ssa_obj.solutions = np.array(solutionssave)
        try:
            ssa_obj.col_points = all_col_points
        except:
            pass
        
        
        
        return ssa_obj
        

 
    def __solve_ssa_lowmem_combined(self,k,t,x0,n_traj,ssa_conditions=None, kon=1, koff=1, kprobe=[] ):
        seeds = np.random.randint(0, 0x7FFFFFF, n_traj, dtype=np.int32)
        
        if isinstance(k,list):
            k = np.array(k).astype(np.float64)
        
        k = k.flatten()

        if kprobe == []:
            kprobe = np.ones(self.color)
            
        if isinstance(kprobe,list):
            kprobe = np.array(kprobe,dtype=np.float64)
        if isinstance(kprobe,int):
            kprobe = np.array([kprobe],dtype=np.float64)
        if isinstance(kprobe,float):
            kprobe = np.array([kprobe],dtype=np.float64)
     
        
        if ssa_conditions == None:
            ssa_conditions = self.default_conditions
        
        x0 = self.__check_x0(x0)
        
        probe_vec = ssa_conditions['probe_vec']
       
        colors = self.colors
        
        
        if ssa_conditions['bursting'] == False:
            kon = 1
            koff = 1
        else:
            kon = kon
            koff = koff
            
        flags = np.array([int(ssa_conditions['bursting']), int(ssa_conditions['leaky_probes']), int(ssa_conditions['record_stats'])],dtype=np.int32)
        probe_loc = ssa_conditions['probe_loc']
        
        
        rib_vec = []
        solutions = []            
        solutionssave = []
        
        
        N_rib = ssa_conditions['ribosome_preallocation']
        all_results,all_nribs,all_collisions,all_frapresults,all_ribtimes,all_col_points = self.__generate_mats_lowmem(n_traj,k[0],t,N_rib,colors)
        footprint = ssa_conditions['footprint']
        evf = ssa_conditions['perturb'][0]
        evi = ssa_conditions['perturb'][1]
        intime = ssa_conditions['perturb'][2]
        non_consider_time = ssa_conditions['burnin']
        
        st = time.time()
        
        for i in range(n_traj):
            
            result,ribtimes,frapresult,coltimes,colpointsx,colpointst = self.__generate_vecs_lowmem(k,t,N_rib,colors)
            nribs = np.array([0],dtype=np.int32)


            if i == 0: #detect any int64
                inputs = [result, ribtimes, coltimes, colpointsx,colpointst, k[1:-1],frapresult, t, k[0], float(k[-1]), int(evf), int(evi), float(intime), seeds[i],nribs,x0,footprint, probe_vec ,int(colors), kon, koff, kprobe, probe_loc, flags,N_rib]
                wash_inputs = self.__check_input_memview(inputs)
            
            if wash_inputs:
                #check memview so all given variables are in int32 if integer for C
                inputs = [result, ribtimes, coltimes, colpointsx,colpointst, k[1:-1],frapresult, t, k[0], float(k[-1]), int(evf), int(evi), float(intime), seeds[i],nribs,x0,footprint, probe_vec ,int(colors), kon, koff, kprobe, probe_loc, flags,N_rib]
                result, ribtimes, coltimes, colpointsx,colpointst, kelong,frapresult, t, ki, kt, evf, evi, intime, seed,nribs,x0,footprint, probe_vec ,colors, kon, koff, kprobe, probe_loc, flags, N_rib = self.__check_memview(inputs)
            else:
                kelong = k[1:-1]
                ki = k[0]
                kt = float(k[-1])
                evf = int(evf)
                evi = int(evi)
                intime = float(intime)
                seed = seeds[i]
                colors = int(colors)
                
            ssa_translation_lowmem.run_SSA(result, ribtimes, coltimes, colpointsx,colpointst, kelong,frapresult, t, ki, kt, evf, evi, intime, seed,nribs,x0,footprint, probe_vec ,colors, kon, koff, kprobe, probe_loc, flags, N_rib)
            #ssa_translation.run_SSA(result, ribtimes, coltimes, k[1:-1],frapresult, truetime, k[0], k[-1], evf, evi, intime, seeds[i],nribs)

            all_results[i, :,:] = result.T
            all_frapresults[i,:] = frapresult
            all_ribtimes[i,:] = ribtimes
            all_collisions[i,:] = coltimes
            all_nribs[i,:] = nribs
            
            endcolrec = np.where(colpointsx == 0)[0][0]
            
            colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
            all_col_points.append(colpoints.T)
                
    
            # for i in range(n_traj):
            #     soln = all_results[i, :].reshape((N_rib, len(t),colors))

            #     so = soln
            #     solutionssave.append(so)
            #     solutions.append(soln)
            
            collisions = np.array([[]])
            watched_ribs = []
            for i in range(n_traj):
                totalrib = all_nribs[i]
            
                if totalrib > all_collisions.shape[1]:
                    collisions = np.append(collisions, all_collisions[i][:])
                    watched_ribs.append(int(all_collisions.shape[1]))
            
                else:
                   
                    collisions = np.append(collisions, all_collisions[i][:int(totalrib[0])])
                    watched_ribs.append(int(totalrib[0]))
            
            sttime = time.time() - st


        no_ribosomes = np.zeros((n_traj, (len(k)+1)))
        
        startindex = np.where(t >= non_consider_time)[0][0]
        
        #all_results = all_results[:,startindex*N_rib:]

#        for i in range(len(solutions)):
#            for j in range(len(solutions[0][0][startindex:])):
#                rib_pos = solutions[i][startindex:, j][np.nonzero(solutions[i][startindex:, j])]
#                print(rib_pos)
#               
#                no_ribosomes[i, rib_pos.astype(int)] += 1
#        no_ribosomes = no_ribosomes[:, 1:]
#
#        ribosome_means = np.mean(no_ribosomes, axis=0)
#        ribosome_density = ribosome_means/len(k)
#
#        no_ribosomes_per_mrna = np.mean(no_ribosomes)
        
        ssa_obj = SSA_Soln()
        ssa_obj.no_ribosomes = no_ribosomes
        ssa_obj.n_traj = n_traj
        ssa_obj.k = k
        #ssa_obj.no_rib_per_mrna = no_ribosomes_per_mrna
        #ssa_obj.rib_density = ribosome_density
        #ssa_obj.rib_means = ribosome_means
        ssa_obj.rib_vec = rib_vec
        ssa_obj.intensity_vec = all_results.T[:,startindex:,:]
        ssa_obj.I = all_results.T[:,startindex:,:]
        ssa_obj.time_vec_fixed = t
        ssa_obj.time = t
        ssa_obj.time_rec = t[startindex:]
        ssa_obj.start_time = non_consider_time
        ssa_obj.watched_ribs = watched_ribs
        ssa_obj.collisions = collisions
        
        
        try:
            ssa_obj.col_points = all_col_points
        except:
            pass
        
        
        ssa_obj.eval_time = sttime
        
        return ssa_obj        

 
    def __solve_ssa_full_combined(self,k,t,x0,n_traj,ssa_conditions=None, kon=1, koff=1, kprobe=[] ):
        seeds = np.random.randint(0, 0x7FFFFFF, n_traj, dtype=np.int32)
        
        if isinstance(k,list):
            k = np.array(k).astype(np.float64)
        
        k = k.flatten()

        if kprobe == []:
            kprobe = np.ones(self.color)
            
        if isinstance(kprobe,list):
            kprobe = np.array(kprobe,dtype=np.float64)
        if isinstance(kprobe,int):
            kprobe = np.array([kprobe],dtype=np.float64)
        if isinstance(kprobe,float):
            kprobe = np.array([kprobe],dtype=np.float64)
     
        
        if ssa_conditions == None:
            ssa_conditions = self.default_conditions
        
        x0 = self.__check_x0(x0)
        
        probe_vec = ssa_conditions['probe_vec']
       
        colors = self.colors
        
        
        if ssa_conditions['bursting'] == False:
            kon = 1
            koff = 1
        else:
            kon = kon
            koff = koff
            
        flags = np.array([int(ssa_conditions['bursting']), int(ssa_conditions['leaky_probes']), int(ssa_conditions['record_stats'])],dtype=np.int32)
        probe_loc = ssa_conditions['probe_loc']
        
        
        rib_vec = []
        solutions = []            
        solutionssave = []
        
        
        N_rib = ssa_conditions['ribosome_preallocation']
        all_results,all_nribs,all_collisions,all_frapresults,all_ribtimes,all_col_points = self.__generate_mats_lowmem(n_traj,k[0],t,N_rib,colors)
        all_rib_loc = np.zeros((n_traj,len(t),N_rib),dtype=np.int32)
        
        
        footprint = ssa_conditions['footprint']
        evf = ssa_conditions['perturb'][0]
        evi = ssa_conditions['perturb'][1]
        intime = ssa_conditions['perturb'][2]
        non_consider_time = ssa_conditions['burnin']
        
        st = time.time()
        
        for i in range(n_traj):

            nribs = np.array([0],dtype=np.int32)
            result,ribtimes,frapresult,coltimes,colpointsx,colpointst = self.__generate_vecs_lowmem(k,t,N_rib,colors)
            ribloc = np.zeros((N_rib,len(t)),dtype=np.int32)   
            if i == 0: #detect any int64
                inputs = [ribloc,result, ribtimes, coltimes, colpointsx,colpointst, k[1:-1],frapresult, t, k[0], float(k[-1]), int(evf), int(evi), float(intime), seeds[i],nribs,x0,footprint, probe_vec ,int(colors), kon, koff, kprobe, probe_loc, flags,N_rib]
                wash_inputs = self.__check_input_memview(inputs)
            
            if wash_inputs:
                #check memview so all given variables are in int32 if integer for C
                inputs = [ribloc,result, ribtimes, coltimes, colpointsx,colpointst, k[1:-1],frapresult, t, k[0], float(k[-1]), int(evf), int(evi), float(intime), seeds[i],nribs,x0,footprint, probe_vec ,int(colors), kon, koff, kprobe, probe_loc, flags,N_rib]
                ribloc,result, ribtimes, coltimes, colpointsx,colpointst, kelong,frapresult, t, ki, kt, evf, evi, intime, seed,nribs,x0,footprint, probe_vec ,colors, kon, koff, kprobe, probe_loc, flags, N_rib = self.__check_memview(inputs)
            else:
                kelong = k[1:-1]
                ki = k[0]
                kt = float(k[-1])
                evf = int(evf)
                evi = int(evi)
                intime = float(intime)
                seed = seeds[i]
                colors = int(colors)
                

            ssa_translation_lowmem.run_SSA_full( ribloc,result, ribtimes, coltimes, colpointsx,colpointst, kelong,frapresult, t, ki, kt, evf, evi, intime, seed,nribs,x0,footprint, probe_vec ,colors, kon, koff, kprobe, probe_loc, flags)
            #ssa_translation.run_SSA(result, ribtimes, coltimes, k[1:-1],frapresult, truetime, k[0], k[-1], evf, evi, intime, seeds[i],nribs)
            
            all_results[i,:, :] = result.T
            all_frapresults[i,:] = frapresult
            all_ribtimes[i,:] = ribtimes
            all_collisions[i,:] = coltimes
            all_nribs[i,:] = nribs
            all_rib_loc[i,:,:] = ribloc.T
            
            endcolrec = np.where(colpointsx == 0)[0][0]
            
            colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
            all_col_points.append(colpoints.T)
            
            
    
            # for i in range(n_traj):
            #     soln = all_results[i,:, :]

            #     so = soln
            #     solutionssave.append(so)
            #     solutions.append(soln)
            
            collisions = np.array([[]])
            watched_ribs = []
            for i in range(n_traj):
                totalrib = all_nribs[i]
            
                if totalrib > all_collisions.shape[1]:
                    collisions = np.append(collisions, all_collisions[i][:])
                    watched_ribs.append(int(all_collisions.shape[1]))
            
                else:
                   
                    collisions = np.append(collisions, all_collisions[i][:int(totalrib[0])])
                    watched_ribs.append(int(totalrib[0]))
            
            sttime = time.time() - st


        no_ribosomes = np.zeros((n_traj, (len(k)+1)))
        
        startindex = np.where(t >= non_consider_time)[0][0]
        
        all_col_points2 = []
        for i in range(len(all_col_points)):
            all_col_points2.append( all_col_points[i][ all_col_points[i][:,1] > non_consider_time ])
        
        
        rib_per_t = np.zeros((n_traj,len(t)))
        for i in range(all_rib_loc.shape[0]):
            rib_loc = all_rib_loc[i,startindex:,:]
            rt = np.count_nonzero(rib_loc.T,axis=0)
            rib_per_t[i,:] = rt
        
        rib_density_loc, rib_density_count = np.unique(all_rib_loc,return_counts=True)
        rib_density_loc = rib_density_loc[1:]
        rib_density_count = rib_density_count[1:]
        rib_density = rib_density_count / np.sum(rib_density_count)
        
        
        ssa_obj = SSA_Soln()
        ssa_obj.no_ribosomes = no_ribosomes
        ssa_obj.n_traj = n_traj
        ssa_obj.k = k
        # ssa_obj.no_rib_per_mrna = no_ribosomes_per_mrna
        ssa_obj.rib_per_t = rib_per_t.T[startindex:,:]
        ssa_obj.rib_mean = np.mean(rib_per_t.T[startindex:,:])

        ssa_obj.rib_density = rib_density
        # ssa_obj.rib_means = ribosome_means
        ssa_obj.rib_vec = rib_vec
        ssa_obj.intensity_vec = all_results.T[:,startindex:,:]
        ssa_obj.I = all_results.T[:,startindex:,:]
        ssa_obj.time_vec_fixed = t
        ssa_obj.time = t
        ssa_obj.time_rec = t[startindex:]
        ssa_obj.start_time = non_consider_time
        ssa_obj.watched_ribs = watched_ribs
        ssa_obj.collisions = collisions
        ssa_obj.ribosome_locations = all_rib_loc
        
        
        try:
            ssa_obj.col_points = all_col_points2
        except:
            pass
        
        
        ssa_obj.eval_time = sttime
        
        return ssa_obj        
 
    
    @classmethod
    def __get_ribosome_statistics(self,ssa_obj,result):
        
        
        return ssa_obj
    
        
    @classmethod
    def __generate_vecs(cls,k,t,N_rib,ncolor):
        tf = t[-1]
        ki = k[0]
        
        guessed_no_ribosomes = int(1.3*ki*tf)
        result = np.zeros((len(t)*N_rib), dtype=np.int32)
        ribtimes = np.zeros((guessed_no_ribosomes),dtype=np.float64)
        frapresult = np.zeros((len(t)*N_rib),dtype=np.int32)
        coltimes = np.zeros((guessed_no_ribosomes),dtype=np.int32)
        colpointsx = np.zeros(len(k[1:-1])*(guessed_no_ribosomes),dtype=np.int32)
        colpointst = np.zeros(len(k[1:-1])*(guessed_no_ribosomes),dtype=np.float64)
        return result,ribtimes,frapresult,coltimes,colpointsx,colpointst

    @classmethod
    def __generate_vecs_trna(cls,kbind,kind,t,N_rib,ncolor):
        tf = t[-1]
        ki = kbind
        
        guessed_no_ribosomes = int(1.3*ki*tf)
        result = np.zeros((len(t)*N_rib), dtype=np.int32)
        ribtimes = np.zeros((guessed_no_ribosomes),dtype=np.float64)
        frapresult = np.zeros((len(t)*N_rib),dtype=np.int32)
        coltimes = np.zeros((guessed_no_ribosomes),dtype=np.int32)
        colpointsx = np.zeros(len(kind)*(guessed_no_ribosomes),dtype=np.int32)
        colpointst = np.zeros(len(kind)*(guessed_no_ribosomes),dtype=np.float64)
        return result,ribtimes,frapresult,coltimes,colpointsx,colpointst
    
    @classmethod
    def __generate_vecs_lowmem(cls,k,t,N_rib,ncolor):
        tf = t[-1]
        ki = k[0]
        
        guessed_no_ribosomes = int(1.3*ki*tf)
        result = np.zeros((ncolor,len(t)), dtype=np.int32)
        ribtimes = np.zeros((guessed_no_ribosomes),dtype=np.float64)
        frapresult = np.zeros((len(t)*N_rib),dtype=np.int32)
        coltimes = np.zeros((guessed_no_ribosomes),dtype=np.int32)
        colpointsx = np.zeros(len(k[1:-1])*(guessed_no_ribosomes),dtype=np.int32)
        colpointst = np.zeros(len(k[1:-1])*(guessed_no_ribosomes),dtype=np.float64)
        return result,ribtimes,frapresult,coltimes,colpointsx,colpointst

    @classmethod
    def __generate_mats_lowmem(cls,ntraj, ki,t,N_rib,ncolor):
        tf = t[-1]
        guessed_no_ribosomes = int(1.3*ki*tf)           
        all_results = np.zeros((ntraj, len(t),ncolor), dtype=np.int32)
        all_ribtimes = np.zeros((ntraj,guessed_no_ribosomes),dtype=np.float64)
        all_frapresults = np.zeros((ntraj,N_rib*len(t)),dtype=np.int32)
        all_collisions = np.zeros((ntraj,guessed_no_ribosomes),dtype=np.int32)
        all_nribs = np.zeros((ntraj,1))
        all_col_points = []
    
        return all_results,all_nribs,all_collisions,all_frapresults,all_ribtimes,all_col_points
        
        
    @classmethod
    def __generate_mats(cls,ntraj, ki,t,N_rib,ncolor):
        tf = t[-1]
        guessed_no_ribosomes = int(1.3*ki*tf)           
        all_results = np.zeros((ntraj, N_rib*len(t)), dtype=np.int32)
        all_ribtimes = np.zeros((ntraj,guessed_no_ribosomes),dtype=np.float64)
        all_frapresults = np.zeros((ntraj,N_rib*len(t)),dtype=np.int32)
        all_collisions = np.zeros((ntraj,guessed_no_ribosomes),dtype=np.int32)
        all_nribs = np.zeros((ntraj,1))
        all_col_points = []
        
        

    
        return all_results,all_nribs,all_collisions,all_frapresults,all_ribtimes,all_col_points
 
    @classmethod
    def __generate_mats_lowmem_nostats(cls,ntraj, ki,t,N_rib,ncolor):

        all_results = np.zeros((ntraj, len(t),ncolor), dtype=np.int32)
        all_frapresults = np.zeros((ntraj,N_rib*len(t)),dtype=np.int32)
        return all_results,all_frapresults
        
    @classmethod
    def __generate_vecs_lowmem_nostats(cls,k,t,N_rib,ncolor):


        result = np.zeros((ncolor,len(t)), dtype=np.int32)

        frapresult = np.zeros((len(t)*N_rib),dtype=np.int32)

        return result,frapresult
    
    def __check_x0(self,x0):
        if len(x0) == 0:
            x0_clean = np.zeros((200),dtype=np.int32)  
        else:
            if len(x0) >200:
                raise ValueError('Unrecognized initial condition, make sure the length of x0 is <= 200')
            
            x0_clean = np.zeros((200),dtype=np.int32)  
            x0_clean[:len(x0)] = x0
        return x0_clean

    def __check_rates_trna(self,rates):
        if isinstance(rates,np.ndarray):
            pass
        else:
            rates = np.array(rates,dtype=np.int32)
        if rates.dtype !=int:
            raise ValueError('trna ID values are not integers, to run the trna ssa version, k_index needs to be an ID of 0-60 integer values')
        if len(np.where(rates < 0)[0]) or len(np.where(rates > 60)[0]):
            raise ValueError('trna ID values are out of bounds, the IDs are 0-60 for tRNA species')      
            
    def __check_input_memview(self,inputs):
        check = False
        for item in inputs:
            try: 
                item.dtype
                if item.dtype not in [np.int32]:
                    if 'int' in str(item.dtype):
                        
                        check = True
            except:
                pass
        return check
    
    def __check_memview(self,inputs):
        #make sure everything is int 32, no int64
        returned_inputs = ()
        for item in inputs:
            try: 
                item.dtype
                if item.dtype not in [np.int32]:
                    if 'int' in str(item.dtype):
                        item = item.astype(np.int32)
                returned_inputs = returned_inputs + (item,)

            except:
                returned_inputs = returned_inputs + (item,)
        return returned_inputs
        
        
    def __check_rates(self,rates):
        if isinstance(rates,np.ndarray):
            pass
        else:
            rates = np.array(rates)
        if len(np.where(rates < 0)[0]) > 0:
            raise ValueError('one or more model rates are negative, double check the provided rates')
       

    def __ssa_python(self, k, t_array, inhibit_time=0, FRAP=False, Inhibitor=False, flags=None, kon=1, koff=1, kprobe=1, ssa_conditions=None):
        '''
        mRNA Translation simulation python implementation

        given a propensity vector k, time array to record, and inhibitory conditions, run a single trajectory of translation simulation

        The simulation is described here: [PUT LINK HERE TO PAPER]

        *args*

            **k**, propensity vector of size gene length + 2, [initiation rate,  Codon dependent rates,  completion rate / unbinding rate]
            for reference the codon dependent rates are refering to the time rate of a ribosome to move on to the next codon

            **t_array**, time points to record the ribosome posistions at

        *keyword args*

            **inhibit_time**, the time to start inhibition assays if FRAP or Inhibitor (harringtonine) == True

            **FRAP**, True or false to apply Fluorescence Recovery After Photobleaching (FRAP) https://en.wikipedia.org/wiki/Fluorescence_recovery_after_photobleaching

            **Inhibitor**, True or false to apply harringtonine at inhibit_time. Harringtonine acts as a protien translation initiation inhibitor

        '''

        #SSA params and propensities
        R = self.default_conditions['footprint'] #exclusion volume (ribosome footprint), ribosomes cant be less than 10 codons apart because of their physical size
        kelong = np.array([k[1:-1]]).T  #rates for ribosomes moving to the next codon, based on tRNA concentrations

        N = len(kelong)  #Number of codons in the mRNA
        kbind = k[0]   #rate for a ribosome to bind and start translation
        kcompl = k[-1]     #rate for a ribosome at the end of the mRNA to unbind
        X = np.array([0, 0], dtype=int)   #the updating ribosome posistion vector that is changed in the simulation
            
        bursting = flags[0]
        leaky = flags[1]
  
        
        if bursting:
            burst = np.random.rand() < (kon/(kon+koff) )
        else:
            burst = 1

        Ncol = np.zeros((1,0))
        
        N_rib = 200  #Maximum number of ribosomes on a single mRNA (hard limit for the simulation not a physical constant)
        

        colors = ssa_conditions['probe_loc'].shape[0]
        if leaky:
            leaky_probe_matrix = np.zeros((colors,N,N_rib))
        intensity = np.zeros((colors,len(t_array)))
        probe = np.array(np.where(ssa_conditions['probe_loc'] ==1))
        probevec = ssa_conditions['probe_vec']
        
            
        #example X arrays and how its formatted:
        # X = [423 30 10 0 ]  read from left to right theres a ribosome in position 423 30 and 10, with a 0 kept as a buffer for simulation

        t = t_array[0]  #time point
        Nt = len(t_array)  #number of time points to record over
        tf = t_array[-1]  #final time point
        col = np.zeros((1,N_rib))
        X_array = np.zeros((N_rib, Nt))  #recording array that records the ribosome posistions over time array points
        NR = 0  #number of ribosomes bound
        it = 1  #number of iterations
        Sn_p = np.eye(max(NR+1, 2), dtype=int) #stoichiometry for the SSA
        if bursting:
            wn_p = np.zeros((X.shape[0]+1, 1)) # propensities for the SSA
        else:
            wn_p = np.zeros((X.shape[0], 1)) # propensities for the SSA
        
        T = np.array([0, 0], dtype=float)
        ribtimes = np.array([[0,0]],dtype=float)
        col_points = []
        #wn_p = np.zeros((1,X.shape[0])).flatten()
        wshape = len(wn_p)
        Inhibit_condition = 1  #set up inhibitor flags
        while t < tf:


            if Inhibitor == True:
                if t >= inhibit_time:

                    Inhibit_condition = 0
                else:

                    Inhibit_condition = 1
            else:
                Inhibit_condition = 1
            if FRAP == True :   #if the Photobleaching is happening, "remove" ribosomes
                if t >= inhibit_time and t < inhibit_time + 20:
                    #X = np.array([0, 0])
                    a=1
                    #T = np.array([0,0])
                    
            oldNR = NR
            NR = len(np.flatnonzero(X)) #each iteration get the number of ribosomes on the mRNA

            if X.shape[0] < NR+1:  #if the last reaction added a ribosome put a 0 on the end of X vec

                X = np.append(X, [0])
                T = np.append(T, [0])
                T[-2] = t
                


            X[-1] = 0
            T[-1] = 0

            X = X[0:max(NR, 1)+1]  #clear any additional 0's on the end
            T = T[0:max(NR, 1)+1]

            if oldNR != NR:     #if the number of ribosomes has changed reallocate the sizes of stoich and propensities
                
                if not bursting:    
                    Sn_p = np.eye(max(NR+1, 2), dtype=int)
                    wn_p = np.zeros((X.shape[0], 1))
                else:
                    Sn_p = np.eye(max(NR+1, 2), dtype=int)
                    wn_p = np.zeros((X.shape[0]+1, 1))        
                    
                if leaky:
                    leaky_probe_matrix[ probe[0] , probe[1] ,NR-1 ] = (np.random.rand(len(probe[0])) < kprobe).astype(int)
                
                    
                wshape = len(wn_p)
                

            Sn = Sn_p
            wn = wn_p

            
            #get indices of where X vecs are > 0 ie where the ribosome values are
            inds = X > 0

            if bursting:
                
                    wn[:-1][inds] = kelong[X[inds]-1]  #update propensities
                
                    
                
            else:
                wn[inds] = kelong[X[inds]-1] 


            if X[0] == N:  #if the ribosome in the 0 position is at the end of the mRNA set propensities to the reaction for completion

                Sn[:, 0] = (np.append(X[1:], np.array([0]))-X[0:])
                

                wn[0] = kcompl


            #if there are no ribosomes or when there is enough room for a new ribosome to bind add the propensity for binding
            if NR == 0:

                wn[NR] = kbind*Inhibit_condition*burst

                

            if NR > 0 and X[NR-1] > R:
                wn[NR] = kbind*Inhibit_condition*burst

            REST = np.less(X[1:]+10, X[0:-1])  #apply the footprint condition ie set any propensities where it violates the > 10 codons apart rule to 0

            if bursting:
                if NR > 1:
                    wn[1:-1] = (wn[1:-1].T*REST).T 
                else:
                    wn[1:] = (wn[1:].T*REST).T 
            else:
                wn[1:] = (wn[1:].T*REST).T  #apply that logical^ to propensities
            
            if bursting:
                if burst:
                    wn[-1] = koff
                else:
                    wn[-1] = kon

            w0 = sum(wn.flat)  #get the sum of propensities
            randnum = np.random.random_sample(2) #update time to point of next reaction (exponential waiting time distb)
            t = (t-np.log(randnum[0])/w0)

            while it < Nt and t > t_array[it]:  #record state if past timepoint
                X_array[0:len(X), it] = X
                
                if leaky:
                    int_tmp = np.zeros(colors)
                    validx = X[X>0]
                    c = leaky_probe_matrix[:,:,0]
                    for ribind in validx:
                        int_tmp += np.sum(c[:,:(ribind-1)],axis=1)
                        
                    intensity[:,it] = int_tmp
                    
                else:
                    validx = X[X>0]
                    int_tmp = np.zeros(colors)
                    for ribind in validx:
                        int_tmp += probevec[:,(ribind-1)]
                    intensity[:,it] = int_tmp
                    
                it += 1
                

     
            if t < tf:  #if still running simulation pick which reaction happened via random number and propensity sum
                r2 = w0*randnum[1]
                tmp = 0
                
                for i in range(wshape):
                    
                    tmp = tmp + wn[i]
                    if tmp >= r2:
                        event = i
                        break
                    

            if bursting:
                if event >= NR+1:
                    burst+=1
                    burst = burst%2
                else:
                    X = (X + Sn[:, event].T)
            
                    if np.sum(Sn[:,event]) < 0 :
                        
                        ribtimes = np.vstack((ribtimes,[T[0],t]))
                        T[:-1] = T[1:]
                        Ncol = np.append(Ncol,col[0][0] )
                        col = np.atleast_2d(np.append(col[:,1:],[0]))
                        
                    else:
                        if X[event-1] == X[event] + R:
                            col[0][event] +=1
                            col_points.append( (X[event],t) )                   

            else:
                X = (X + Sn[:, event].T)  #update X vector for new ribosome state
            
            
                if np.sum(Sn[:,event]) < 0 :
                    
                    ribtimes = np.vstack((ribtimes,[T[0],t]))
                    T[:-1] = T[1:]
                    Ncol = np.append(Ncol,col[0][0] )
                    col = np.atleast_2d(np.append(col[:,1:],[0]))
                    
                    if leaky: #remove the leaky probes
                        leaky_probe_matrix[ probe[0] , probe[1] ,0 ] = leaky_probe_matrix[ probe[0] , probe[1] ,1 ]
                
                    
                else:
                    if X[event-1] == X[event] + R:
                        col[0][event] +=1
                        col_points.append( (X[event],t) )
                    
                
            
        return X_array,ribtimes[1:,:],Ncol,col_points,intensity  #return the completed simulation




# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 17:52:28 2020

@author: willi
"""
import numpy as np
from scipy.sparse.linalg import expm
from scipy.integrate import odeint
from scipy.integrate import ode
import scipy as sp
import scipy.optimize
import random as random
import collections as c
#import matplotlib.pyplot as plt
import copy as cp
from multiprocessing import process as Process
#import matplotlib.pyplot as plt

class GenericSSA:
    def __init__(self,type='linear'):
        self.xi=np.array([])
        self.ti=None
        self.tf=None
        self.S=np.array([])
        self.type=type
        self.ptimes=100
        self.params={}
        if type=='linear':
            self.W0=np.array([])   
            self.W1=np.array([])
        if type == 'nonlinear':
            self.P=lambda x,t:None 
    def gettvec(self):
        return np.linspace(self.ti,self.tf,self.ptimes)
        
    def updateparams(self):
        for k, v in self.params.iteritems():
            setattr(self, k, v)

    def _run_trajectory(self):#!!!!!!!!!!!!!!!!!renamed run to solve(big deal)
        x=self.xi
        t=0
        __n=len(x)
        self.time=self.gettvec()
        data=np.zeros((len(self.xi),self.ptimes))
        ip=0
        if self.type=='linear':
            while t<self.tf:
                rate=np.atleast_2d(np.dot(self.W1,x))+self.W0
                rate=np.cumsum(rate)
                t=(t-np.log(np.random.rand(1))/rate[-1])
                ro=rate[-1]*np.random.rand()
                while t>self.time[ip]:
                    if t>self.tf:
                        b = len(self.time[ip:])
                        fill = np.repeat(x,b)
                        data[:,ip:]=fill.reshape(__n,b)
                        return data
                    else:
                        data[:,ip]=x.reshape(__n)
                        ip=ip+1
                for i in range(len(rate)):
                    if rate[i]>=ro:
                        event=i
                        break
                x=x+np.atleast_2d(self.S[:,event]).T
        elif self.type=='nonlinear':
            pass
        else:
            'Error'
        return data

    def _solve(self,n):
        soln=np.zeros((len(self.xi),self.ptimes,n))
        for i in range(n):
            d=self._run_trajectory()
            soln[:,:,i]=d
        self.soln = soln
        return soln
        
    def getW0(self):
        pass
    
    def getW1(self):
        pass
    
    def setpar(self,key,val):
        self.params[key]=val

class GenericFSP:
    def __init__(self,ti=[],tf = [],xi=[],A=np.array([]),ptimes=100):
        self.ti=ti
        self.tf=tf
        self.xi=xi
        self.A=A
        self.errors=np.empty(ptimes)
        self.ptimes=ptimes
        self.params={}
        
    def updateparams(self):
        for k, v in self.params.iteritems():
            setattr(self, k, v)
        
    def gettvec(self):
        '''
        get vector of times 
        '''
        return np.linspace(0,self.tf-self.ti,self.ptimes)

    def _solve(self):
        '''
        solve the FSP at each time. 
        '''
        tvec=self.gettvec()
        self.model_solns = np.zeros((len(self.xi),len(tvec)))
        for i in range(len(tvec)):
            mexp = expm(self.A*tvec[i])
            self.soln = mexp.dot(self.pi)
#           self.errors[i]=1.0-np.sum(self.model_solns[:,i])
        return self.soln

class GenericODE:
    def __init__(self,ti=None,tf=None,ODE=None,ptimes=50,xi=None):
        self.ti=ti
        self.tf=tf
        self.ODE=ODE
        self.ptimes=ptimes
        self.xi=xi
        self.tvec=None
        self.soln=None
        self.parameters={}
    
    def updateparams(self):
        try:
            for k, v in self.params.items():
                setattr(self, k, v)
        except:
             for k, v in self.params.iteritems():
                setattr(self, k, v)           
            
    def gettvec(self):
        '''
        get vector of times
        '''
        return np.linspace(self.ti,self.tf,self.ptimes)

    def _solve(self):
        '''
        solve the ODEs at each time
        '''
        self.tvec=self.gettvec()
        try: 
            solution=odeint(self.ODE,self.xi,self.tvec,Dfun=self.Jacobian)
        except: 
            solution=odeint(self.ODE,self.xi,self.tvec)
        self.soln = solution.T
        return solution.T
        
    def _expm_solve(self):
        self.tvec = self.gettvec()
        
        
        
class GenericStats():
    '''
    a class of static functions to analyze data
    '''
    def __init__(self,model=None):
        if model is not None:
            self.model=model
            self.type=model.__class__.__name__
            self.simdata={}
            self.simdata['data']=self.model.solve()
            self.simdata['time']=self.model.gettvec()
            
            self.means=self.get_means(**self.simdata)
            self.acc=self.get_acc(self.simdata)
            if self.__class__.__bases__ is 'GenericSSA' or self.__class__.__name__ is 'GenericSSA':
                self.var=self.get_var(self.simdata['data'])

    @staticmethod
    def get_var(data,time):
        '''
        a function that returns statistical variance of data
        '''
        if type(data) is not np.ndarray :
            print("Error: data is not a numpy array")
            return
        s=data.shape
        if len(s) is 1:
            s=s+1
        var=np.zeros((s[0],s[1]))
        for j in range(len(time)):
            for i in range(s[0]):
                var[i,j]=np.var(data[i,j,:])
        return var
        
    @staticmethod
    def get_means(data,time):
        '''
        a function that returns the statistical mean of data
        '''
        if type(data) is not np.ndarray :
            print("Error: data is not a numpy array")
            return
        __s=data.shape
        means=np.zeros((__s[0],__s[1]))
        for j in range(len(time)):
            for i in range(__s[0]):
                means[i,j]=np.mean(data[i,j,:])
        return means
    
    @staticmethod 
    def get_acc2(data,trunc=True): 
        N = len(data)    
        fvi = np.fft.fft(data,n=2*N)     
        acf = fvi*np.conjugate(fvi)
        acf = np.fft.ifft(acf)
        acf = np.real(acf[:N])/float(N)
        if trunc:
            acf[acf<0]=0
            for i in range(1,len(acf)): 
                if acf[i] > acf[i-1]:
                    acf[i] = acf[i-1]
        return acf
        
    @staticmethod 
    def get_fastacc2(data,trunc=True): 
        N = len(data)    
        fvi = np.fft.fft(data,n=2*N)     
        acf = fvi*np.conjugate(fvi)
        acf = np.fft.ifft(acf)
        acf = np.real(acf[:N])/float(N)
        if trunc:
            acf[acf<0]=0
            for i in range(1,len(acf)): 
                if acf[i] > acf[i-1]:
                    acf[i] = acf[i-1]
        return acf
    @staticmethod
    def get_acc(data,one_sided=True,trunc=True):
        '''
        a function that return the statistical autocorreletion of data
        '''
        half = len(data)/2
        z = np.correlate(data[:half+1],data,mode='valid') / float(half)
        if trunc:
            z[z<0]=0
            for i in range(1,len(z)): 
                if z[i] > z[i-1]:
                    z[i] = z[i-1]
        return z 

    @staticmethod
    def get_xcc(data1,data2=None,axis=1,norm=False,trunc=False):
        '''
        a function that returns the statistical crosscorrelation(or autocorrelation)of data
        '''
        if data2 is None:
            data2=data1
        if type(data1) is not np.ndarray:
            return "Type error, data must be a numpy array"
        corr = np.correlate(data1,data2,mode='full')
        if norm and corr.max()>0:
            corr = corr/corr.max()
        lc = len(corr)/2
        if trunc:
            corr[corr<0]=0
            tmp = corr[lc:]
            for i in range(1,lc): 
                if tmp[i] > tmp[i-1]:
                    tmp[i] = tmp[i-1]
            corr[lc:] = tmp
        return corr 

    @staticmethod
    def get_kld(p,q):
        '''
        returns the kullback-leibler divergence between the distributions 
        p and q. 
        '''
        support = (p>0)*(q>0)
        #if np.max(self.target)<np.min(self.predicted) or np.min(self.target)>np.max(self.predicted):
        #    kld = np.inf
        kld = np.sum( p[support] * (np.log(p[support]) - np.log(q[support])))
        return kld    
    
    @staticmethod
    def normalize(data):
        return data/sum(data)
        
    @staticmethod
    def get_curvature(y,h):
        k=np.zeros(len(y)-2)
        for i in range(len(y)-2):
            k[i]=((y[i+2]-2*y[i+1]+y[i])/(h**2.0))/(1+((y[i+2]-y[i])/(2*h))**2)**(1.5)
        return k
        
    @staticmethod
    def get_sse(data1,data2,axis=None):
        '''calculates sum of squared error of two numpy array datasets'''
        return np.linalg.norm(data1-data2)
        
class GenericOpt:
    def __init__(self,model):
        self.model=model
        self.modeltype=model.__class__.__name__
        self.activeobjfun='chisq'
        self.activemethod='MH'
        self.activefun='solve'
        self._mdt={'met_hast':self.methast,\
                      'methast':self.methast,\
                      'MH':self.methast,\
                      'lin_opt':self.lin_opt,\
                      'linopt':self.lin_opt,\
                      'genetic':self.genetic,\
                      'sim_anneal':self.sim_anneal,\
                      'simanneal':self.sim_anneal}
        self._odt={'sse':self.sse,\
                   'SSE':self.sse,\
                   'chi':self.chisq,\
                   'chisq':self.chisq}
        self.opts={'bounds':([0,10],)*len(self.model.params)}
        self.data=[]
        self.time=[]
        self.cache=c.deque([])
        self.bestpars=[]
    def fit(self,*args):
        #if method is not None:
        #    return self.methods[method](self,args)
        if self.activemethod is not None and self.activeobjfun is not None:
            print("Running: " +self.activemethod +" with "+self.activeobjfun)
            return self._mdt[self.activemethod](*args)
        else:
            print("Please pick an active solver and objective function")
            
    def getsimdata(self):
        return self.model.solve()
    
    def sse(self):
        simdata=self.getsimdata()
        return np.sum((self.data-simdata)**2)

    def chisq(self):
        simdata=self.getsimdata()
        return np.sum(((self.data-simdata)**2)/self.model.variance)
    
    def kld(self,x):
        pass#return np.stats.kld(1,2,3)
    def methast(self,N=1000,burn=100,stepsize=5):
        def evolvepars(p,stepsize):
            new=cp.copy(p)
            for j in range(len(new)):
                while 1 :
                    new[j]=p[j]+np.random.normal(0,stepsize)
                    #print p
                    #print new
                    if (self.opts['bounds'][j][0]<new[j]<self.opts['bounds'][j][1]):
                        break
            return new
                        
        optfun=self.getoptfun()
        oldpars=self.bestpars=self.getpars()
        f_old=f_best=optfun(oldpars)   
        for i in range(-burn,N):
            newpars=evolvepars(oldpars,stepsize)
            f_new=optfun(newpars)
            #print("newpars = "+str(newpars)+ "     oldpars = "+str(oldpars))
            #print("fnew="+str(f_new)+", fold="+str(f_old))
            if f_new<f_old or np.random.rand()<np.exp(f_old-f_new):
                #raw_input("Press Enter to continue...")
                f_old=f_new
                oldpars=newpars
                print("Accpeted "+str(i))
                if i>0:
                    self.cacheitem(newpars)    
                if f_new<f_best:
                    f_best=f_new
                    self.bestpars=newpars
                

        return self.bestpars
    
    def genetic(self):
        f=self.getoptfun()
        logfun = lambda x: f(x)
        return sp.optimize.differential_evolution(logfun,**self.opts)
    def lin_opt(self):
        pass
    def sim_anneal(self):
        return sp.optimize.minimize(self.getoptfun(),self.getpars(),method='Anneal',options={'lower':[0,0]})
    
    def getoptfun(self):
        def fun(pars):
            self.setpars(pars)
            #print self.model.params
            return self._odt[self.activeobjfun]()
        return fun
        
    def getpars(self):
        i=0
        pars=np.zeros(len(self.model.params))
        for key,val in self.model.params.iteritems():
                pars[i]=val
                i+=1
        return pars
    def clearcache(self):
        self.cache=c.deque([])
    def cachepars(self):
        self.cache.append((self.getpars()))
    def cacheitem(self,item):
        self.cache.append(item)
    def setpars(self,parval):
        i=0
        for key,val in self.model.params.iteritems():
                self.model.params[key]=parval[i]
                i+=1
    def parplot(self,s=2,alpha=.5):
        N=len(self.model.params)
        D=zip(*self.cache)
        #print D

        for i in range(N):
            for j in range(N):
                plt.subplot(N,N,j+N*i+1)
                if i is not j:                    
                    plt.scatter(D[i],D[j],linewidth=0.0,s=s,alpha=alpha)
                else:
                    plt.hist(D[i])
        
class GenericModel(GenericSSA,GenericODE):
    '''
    define a generic model to run SSA,ODE~may add FSP analysis here one day
    '''
    def __init__(self,ti=None,tf=None,ODE=None,ptimes=50,xi=None):
        self.ti=ti
        self.tf=tf
        self.ODE=ODE
        self.ptimes=ptimes
        self.xi=xi
        self.tvec=None
        self.soln=None
        self.parameters={}
        self.type=None
    def updateparams(self):
        for k, v in self.params.iteritems():
            setattr(self, k, v)
            
    def gettvec(self):
        '''
        get vector of times
        '''
        return np.linspace(self.ti,self.tf,self.ptimes)

    def _solve(self):
        '''
        solve the ODEs at each time
        '''
        self.tvec=self.gettvec()
        try: 
            solution=odeint(self.ODE,self.xi,self.tvec,Dfun=self.Jacobian)
        except: 
            solution=odeint(self.ODE,self.xi,self.tvec)
        self.soln = solution.T
        return solution.T
    



# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 17:34:04 2020

@author: willi
"""
import numpy as np
from . import IntensityAnalyses
from . import PropensityFactory, ProbeVectorFactory
import scipy as sci
import copy
import time
import pandas as pd

import pickle
import matplotlib.pyplot as plt
from matplotlib import cm
from scipy.stats import multivariate_normal,chi2
from matplotlib.patches import Ellipse

class TranslationOptimization():
    '''
    optimization class adjusted from Micheal May's GenericOpt
    '''        
    def __init__(self):
        
        self.solver_obj = None
        self.data_obj = None
        self.int_a = IntensityAnalyses()
        
        self.parnames = []
        self.params = np.array([])
        self.initial_params = np.array([])
        
        self.pff = PropensityFactory()
        self.pvf = ProbeVectorFactory()
        
        
        self._int_a = IntensityAnalyses()
        self.methods={'met_hast':self.methast,\
                      'methast':self.methast,\
                      'MH':self.methast,\
                      'lin_opt':self.lin_opt,\
                      'linopt':self.lin_opt,\
                      'genetic':self.genetic,\
                      'sim_anneal':self.sim_anneal,\
                      'simanneal':self.sim_anneal}
            
        self.objective_funs={'sse':self.sse,\
                   'SSE':self.sse,\
                   'chi':self.chisq,\
                   'chisq':self.chisq,\
                   'LL_acorr':self.get_loglikelihood_autocorr,\
                   'I_mu_sse':self.get_intensity_sse,\
                   'LL_I_distb':self.get_loglikelihood_intensity_distribution,\
                   'LL_acorr_ode':self.get_loglikelihood_autocorr_ode}
            
        self.opts={'bounds':([0,10],)*len(self.params)}  
        
        self.args = {'LL_acorr': (20,'raw','G0'), 'I_mu_sse':(), 'combined_objective':[],'LL_I_distb':(30,True)}
        
        self._obj_weights = self.args.copy()
        for key in self._obj_weights.keys():
            self._obj_weights[key] = 1
        
     
        
        self.last_run = None
        self.chain = None
        
    def print_args(objfun):
        x=1
        
    
    def add_bounds(self,lower_bounds, upper_bounds):
        '''
        Adds bounds to the optimizer object

        Parameters
        ----------
        lower_bounds : List or ndarray
            lower bound of each parameter.
        upper_bounds : List or ndarray
            upper bound of each parameter.

        Returns
        -------
        None.

        '''
        if isinstance(lower_bounds,np.ndarray):
            lower_bounds = lower_bounds.tolist()
        if isinstance(upper_bounds,np.ndarray):
            upper_bounds = upper_bounds.tolist() 
            
        self.opts['bounds'] = tuple(list(x) for x in zip(lower_bounds,upper_bounds))
        
        
    def intensity_fun(self,x):
        self.solver_obj._poi.ke_mu = x[1]
        self.solver_obj._poi.ki = x[0]
        ssa_soln = self.solver_obj.solve_ssa_set_conditions()
        return ssa_soln.intensity_vec
    
    def autocovariance_fun(self,intensity,norm='ind'):
        acov,err_acov = self._int_a.get_autocov(intensity,norm=norm)        
        return acov,err_acov

    def autocorrelation_fun(self,intensity,norm='ind',g0='G0'):
        acov,err_acov = self._int_a.get_autocov(intensity,norm=norm)
        acorr,err_acorr = self._int_a.get_autocorr(acov,g0=g0)
        return acorr,err_acorr    

    def intensity_distribution(self,intensity,bins=None,density=True,norm=1):
        
       
        int_dist = np.histogram(intensity/norm, bins=bins, density=density,) 
        int_dist_bins = int_dist[1]
        int_dist_heights = int_dist[0]
        return int_dist_heights,int_dist_bins
    
    def analytical_autocorrelation(self,x,bins=None,bin_method='intellegent'):
        
        self.solver_obj._poi.ke_mu = x[1]
        self.solver_obj._poi.ki = x[0]       
        
        if not isinstance(bins,None):
            
            
            inds = self.pff.intellegent_bin(np.atleast_2d(self.solver_obj._poi.probe_loc),100)
            bpl,bpv = self.pvf.bin_probe_vecs(self.solver_obj._poi.probe_loc,inds)
            k_bin = self.pff.bin_k(self.solver_obj._poi.kelong, inds)
            x0 = np.zeros((k_bin.shape[0],1))
            t = self.solver_obj.t
            ode_soln_bin = self.solver_obj.solve_ode(k_bin,   t, x0, self.solver_obj._poi.ki, bpl,corr=True)
        
        
    def genetic(self,objfun,**kwargs):
        
        return sci.optimize.differential_evolution(objfun,self.opts['bounds'],**kwargs)
    
    def sim_anneal(self,objfun,**kwargs):
        
        minimizer_kwargs = {"bounds":self.opts['bounds'] }
        return sci.optimize.basinhopping(objfun,self.initial_params,minimizer_kwargs =minimizer_kwargs,**kwargs)
    
    
    def methast(self, optfun, optfun_type, niter=1000, burnin=100, stepsize=None,mut_rate=.3, disp=False, logspace=True, proposal = None):
        '''

        Parameters
        ----------
        optfun : function
            objective function to optimize.
        optfun_type : str
            the name of the objective function for __.args to access its arguments
        niter : int, optional
            Number of iterations. The default is 1000.
        burnin : int, optional
            Number of iterations to burn (disregard from the final chain). The default is 100.
        stepsize : list, ndarray, optional
            The stepsize of the proposal distribution. The default is None.
        mut_rate : float 0-1, optional
            mutation probability. The default is .3.
        disp : boolean, optional
            return a display of the current progress. The default is False.
        logspace : boolean, optional
            use logspace to search. The default is True.
        proposal : function, optional
            function to draw proposal distributions from. The default is None.

        Returns
        -------
        OptimizeResult
            scipy result object.

        '''
        if logspace:
            bounds = tuple([np.log10(x).tolist() for x in self.opts['bounds']])
            initial_par = np.log10(self.initial_params)
        else:
            bounds = self.opts['bounds']
            initial_par = self.initial_params
        
        if stepsize is None:
            stepsize = (initial_par/10).tolist()
            
        if proposal is None:
            proposal = np.random.normal
            
        
        def evolvepars(p,stepsize):
            new=np.copy(p)
            for j in range(len(new)):
                    if np.random.rand() < mut_rate:
                        while 1:
                            new[j]=p[j]+proposal(0,stepsize[j])
                    #print p
                    #print new
                            if (bounds[j][0]<new[j]<bounds[j][1]):
                                break
            return new
                        
        objective_fun_args = self.args[optfun_type]
        

        if logspace:
            oldpars=initial_par
            f_old=f_best=optfun(10**oldpars, *objective_fun_args)             
            bestpars=initial_par
        else:
            oldpars=initial_par
            f_old=f_best=optfun(oldpars, *objective_fun_args) 
            bestpars=initial_par
            
        if np.isnan(f_old):
            f_old = np.inf
        
        
    
        if disp:
            print('Burning in....')
        
        for i in range(-burnin,niter):
            if disp:
                if i > 0 and i%(niter/10) == 0:
                    self.__mh_print_report(i,bestpars,f_best,niter)
            
            newpars=evolvepars(oldpars,stepsize)

            if logspace:
                tmp_pars = copy.deepcopy(newpars)
                f_new=optfun(10**tmp_pars,*objective_fun_args)
            else:
                f_new=optfun(newpars,*objective_fun_args)
            #print("newpars = "+str(newpars)+ "     oldpars = "+str(oldpars))
            #print("fnew="+str(f_new)+", fold="+str(f_old))
            
            if np.isnan(f_new):
                f_new = np.inf
     
            if f_new<f_old or np.random.rand()<np.exp(f_old-f_new):

                #raw_input("Press Enter to continue...")
                f_old=f_new
                oldpars=newpars
                self.__update_mh_chain(newpars,f_new)
                if f_new<f_best:
                    f_best=f_new
                    bestpars=newpars
                    
        result = sci.optimize.OptimizeResult()
        
        if logspace:
            result.x = 10**bestpars
        else:
            result.x = bestpars
        
        result.fun = f_best
        result.success = True
        result.nit = niter
        result.logspace = logspace
        return result
    
    def __mh_print_report(self,iteration,bestpars,fbest,niter):
        print('current iteration: %d out of %d | best_parameters: %s | best evaulation: %f' % (iteration,niter,''.join(str(bestpars.tolist())),fbest )  )
    
    def __update_mh_chain(self,pars,funeval):
       
        self.chain.accepted_parchain = np.vstack( (self.chain.accepted_parchain, pars) )
        self.chain.accepted_evalchain = np.vstack((self.chain.accepted_evalchain,np.sum(funeval)))
        #self.chain.accepted_objfunchain  = np.vstack((self.chain.accepted_objfunchain,np.atleast_1d(funeval)))
        self.chain.accepted = self.chain.accepted + 1
        

            
    def lin_opt(self):
        return 1
    
    
    def combined_objective(self,x,objfun_list,intensity_fun):
        obj_sum = 0
        intensity = intensity_fun(x)
        obj_fun_evals = np.zeros(len(objfun_list))
        k = 0
        for i in range(len(objfun_list)):
            objective = objfun_list[i]
            obj_args = self.args[objective]
            
            obj_fun_evals[i] = self._obj_weights[objective]*self.objective_funs[objective](intensity,*obj_args)
         
            #obj_sum += self._obj_weights[objective]*self.objective_funs[objective](intensity,*obj_args)  
        obj_sum = np.sum(obj_fun_evals)
        self.__update_chain(x, obj_fun_evals )
        
        return obj_sum
    
    def run_optimization(self, objective_fun_list, method ,model = None, data = None, intensity_fun = None,**kwargs):
        if isinstance(objective_fun_list,str):
            objective_fun_list = [objective_fun_list]
        
        if model == None:
            model = self.solver_obj
        if data == None:
            data = self.data_obj
        if intensity_fun == None:
            intensity_fun = self.intensity_fun
                        
        obj_fun = self.combined_objective
                
        method_fun = self.methods[method]
        
        self.chain = OptChain()
        
        self.chain.parchain = self.initial_params
        self.chain.initial_params = self.initial_params
        self.chain.iterations = 0
        self.chain.parnames = self.parnames
        
        self.chain.bestpar = None
        self.chain.besteval = None
        self.chain.opt_method = method
        

        self.chain.evalchain  = np.array([11110])
        self.chain.objfunchain = np.zeros((1,len(objective_fun_list)))
        
        starttime = time.time()
        
        args = (objective_fun_list,intensity_fun)
        
        if method in ['met_haste','methaste','MH']:
            self.args['combined_objective'] = [objective_fun_list,intensity_fun]
            self.chain.accepted = 0
            self.chain.accepted_evalchain = np.array([11110])
            self.chain.accepted_objfunchain =  np.zeros((1,len(objective_fun_list)))
            self.chain.accepted_parchain = self.initial_params
            result = method_fun(obj_fun,'combined_objective', **kwargs)
            self.chain.accepted_evalchain = self.chain.accepted_evalchain[1:]
            #self.chain.accepted_objfunchain =  self.chain.accepted_objfunchain[1:,:]
            self.chain.accepted_parchain = self.chain.accepted_parchain[1:,:]     
            
        else:
            kwargs['args'] = args
            result =  method_fun(obj_fun,**kwargs)
            
        self.chain.runtime = time.time()-starttime
        
        #result =  method_fun(obj_fun,**kwargs)
        
        self.chain.bestpar = result.x
        self.chain.besteval = result.fun
        
        self.chain.parchain = self.chain.parchain[1:,:]
        self.chain.objfunchain = self.chain.objfunchain[1:,:]
        self.chain.evalchain  = self.chain.evalchain[1:]
        

    def __update_chain(self,pars, funeval):   

  
        self.chain.parchain = np.vstack( (self.chain.parchain, pars) )

        self.chain.evalchain = np.vstack((self.chain.evalchain,np.sum(funeval)))
        self.chain.objfunchain  = np.vstack((self.chain.objfunchain,np.atleast_1d(funeval)))
        self.chain.iterations = self.chain.iterations + 1
        
        
    def __intensity_generator(self,pars,objective_fun_list):
        intensity = self.intensity_fun(pars)
        
        return intensity
    
    def get_loglikelihood_autocorr(self, intensity, n_points,norm,g0):
        
        model_acorr,model_acorr_err = self.autocorrelation_fun(intensity,norm=norm,g0=g0)

        total_n_spots = self.solver_obj.n_traj
        data_autocorrelation = self.data_obj.acorr
        data_acc_err = self.data_obj.acorr_err
        LL = self.loglikelihood_acc(model_acorr[:,:n_points,:],  data_autocorrelation[:,:n_points,:], data_acc_err[:,:n_points], total_n_spots)
    
        
        
        return LL

    def get_loglikelihood_autocorr_ode(self, intensity, n_points,norm,g0):
        
        model_acorr,model_acorr_err = self.autocorrelation_fun(intensity,norm=norm,g0=g0)

        total_n_spots = self.solver_obj.n_traj
        data_autocorrelation = self.data_obj.acorr
        data_acc_err = self.data_obj.acorr_err
        LL = self.loglikelihood_acc(model_acorr[:,:n_points,:],  data_autocorrelation[:,:n_points,:], data_acc_err[:,:n_points], total_n_spots)
    
       
        
        return LL

    def get_intensity_sse(self,intensity):
        
        return (np.mean(intensity) - np.mean(self.data_obj.I_mu))**2
    
    def get_loglikelihood_intensity_distribution(self,intensity,norm):
        
        dist_sim_data = self.intensity_distribution(intensity,bins=self.data_obj.histogram_bins,density=True,norm=norm)[0]
        dist_sim_data[dist_sim_data==0] = 1e-7
        
        LL = -np.dot(self.data_obj.histogram,np.log(dist_sim_data))
        
        if LL == -np.inf:
            LL = np.inf
        return LL
        
    @staticmethod
    def sse(model_data,data):        
        return np.sum((model_data-data)**2)
    
    @staticmethod
    def chisq(model_data,model_var, data):
        return np.sum(((model_data-data)**2)/model_var)
    
    
    @staticmethod
    def loglikelihood_acc( model, data,data_err, nspots):
        '''
        Parameters
        ----------
        model : ndarray
            autocorrelation model generated data array.
        data : ndarray
            autocorrelation data array.
        data_err : ndarray
            error array (SEM, STD) of the data.
        nspots : int
            number of spots.

        Returns
        -------
        Loglikelihood (float)
            returns the logliklihood from the formula:
                
                log L(G|M) = Const. - 1/Nt * sum( (G_D(t_i) - G_M(t_i))^2 / sigma_G_D(t_i) )

        '''
        
        d = np.mean(data,axis=-1)
        m = np.mean(model,axis=-1)
       

    
        return (nspots/2) * np.sum(( d[:,2:] - m[:,2:])**2/ data_err[:,2:])
    
    @staticmethod
    def loglikelihood_distb( model_intensity, data_intensity,nbins=30,norm=1):
        '''
    
        Parameters
        ----------
        model_intensity : ndarray
            array of model simulated intensity over time.
        data_intensity : ndarray
            array of data intensity over time.
        nbins : int, optional
            Number of bins. The default is 20.

        Returns
        -------
        float
            LogLikelihood of the intensity distributions.

        '''
        dist_sim_data = np.histogram(model_intensity/norm, bins=nbins, density=True)[1]
        hist_exp_data = np.histogram(data_intensity, bins=nbins)
        return -np.dot(hist_exp_data,np.log(dist_sim_data))
    

class IntensityData():
    def __init__(self):
      
        self.ragged = False
        self.head = None
        self.ssa_obj = None    
        
    
    def add_data(self, t, intensity_vec):
        '''

        Parameters
        ----------
        intensity_vec : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        '''
        self.intensity_vec = intensity_vec
        if isinstance(intensity_vec,np.ndarray):
            self.ragged=False
            
        else:
            self.ragged=True
        self.times = t
        
         
    def get_stats(self):
        if self.ragged == False:
            self.I_mu = np.mean(self.intensity_vec,axis=2)
            self.I_var = np.mean(self.intensity_vec,axis=2)
        else:
            spot_no = len(self.intensity_vec)
            def autocorr(x):
                result = np.correlate(x, x, mode='full')
                return result[result.size // 2:]        
            # Calculating the autocovariance
            n_lags = 300
            n_selected_Particles = np.max(spot_no)
            ac_array = np.zeros((n_selected_Particles , n_lags ))
            norm_ac_array = np.zeros((n_selected_Particles , n_lags ))
            counter = 0
            # Calculating Autocorrelation.
            for i in range(1,n_selected_Particles+1):
                intensity_green = self.intensity_vec[i][0]
                temp_ac = autocorr(intensity_green)
                size_ac_temp  = len(temp_ac)
                ac_array[counter, 0:size_ac_temp] = autocorr(intensity_green)
                norm_ac_array[counter, 0:size_ac_temp] = autocorr(intensity_green)/ float(ac_array[counter, :] .max())
                counter += 1
            # Plotting mean autocovariance
            lag_time = [*range(0, self.max_lag_output, 1)]
            lag_time = [element * self.sampling_rate for element in lag_time]
            mean_ac_data = norm_ac_array.mean(0)
            std_ac_data = norm_ac_array.std(0)
            # normalized autocovariance, removing shot noise
            
            if self.remove_shotnoise ==1:
                #mean_ac_data_norm = mean_ac_data[1:-1]/mean_ac_data[1]
                return lag_time[1:self.max_lag_output], mean_ac_data[1:self.max_lag_output], std_ac_data[1:self.max_lag_output]
            else:
                #mean_ac_data_norm = mean_ac_data[1:-1]/mean_ac_data[0]
                return lag_time[0:self.max_lag_output], mean_ac_data[0:self.max_lag_output], std_ac_data[0:self.max_lag_output]
            
        
    def load_data(self, file):
        extension = file.split('.')[-1]
        if extension in ['xls', 'xlsx']:
            df = pd.read_excel(file)
            
            self.head = df.head()
            n_cells = max(df['Cell_#'])
            max_spots =  max(df['Spot_#'])
            total_n_spots = 0
            data_struct = []
            for i in range(0,n_cells):
                i_g = []
                i_r = []
                intensitys = []
                times = []
                for j in range(0,max_spots):
                    temp_len_spot =  df[(df['Cell_#'] == i) & (df['Spot_#'] ==j) ]
                    if len(temp_len_spot) >0:
                        time = df[(df['Cell_#'] ==i) & (df['Spot_#'] ==j)]['Time_(sec)'].values
                        intensity_red = df[(df['Cell_#'] ==i) & (df['Spot_#'] ==j)]['R_GaussianInt'].values
                        intensity_green = df[(df['Cell_#'] ==i) & (df['Spot_#'] ==j)]['G_GaussianInt'].values
            
                        i_g.append(intensity_green)
                        i_r.append(intensity_red)
                        
                        intensity = np.vstack((intensity_red,intensity_green))
                        intensitys.append(intensity)
                        times.append(time)
                data_struct.append([i_g,i_r,intensitys,times ])
                
            time_arrays = []
            intensity_arrays = []
            cells = []
            for n in range(len(data_struct)):
                cells = cells + [n for x in range( len(data_struct[n][3])  )]
                time_arrays = time_arrays + data_struct[n][3]
            
            for n in range(len(data_struct)):
                intensity_arrays = intensity_arrays + data_struct[n][2]
            
            len_t = len(time_arrays[0])
            self.ragged = False
            for n in range(len(time_arrays)):
                if len(time_arrays[n]) != len_t:
                    self.ragged = True
                    
            self.intensity_vec = intensitys
            self.times = time_arrays
            self.cells = cells
            
            
    def __guess_headers(self,dataframe):
        columns = dataframe.columns.tolist()
        
        cell_header = None; spot_header = None; time_header = None;
        R_header=None; G_header = None; B_header=None;
    
        kwargs = {}
        for col in columns:
            if 'spot' in col.lower():
                spot_header = col
                kwargs['spot_header'] = spot_header
            if np.sum([x in col.lower() for x in ['green','g_']]) >0:
                green_header = col
                kwargs['green_header'] = green_header
            if np.sum([x in col.lower() for x in ['red','r_']]) >0:
                red_header = col
                kwargs['red_header'] = red_header
            if np.sum([x in col.lower() for x in ['blue','b_']]) >0:
                blue_header = col  
                kwargs['blue_header'] = blue_header
            if 'cell' in col.lower():
                cell_header = col
                kwargs['cell_header'] = cell_header
            if np.sum([x in col.lower() for x in ['time','sec']]) >0:
                time_header = col   
                kwargs['time_header'] = time_header
        
        return kwargs
        
    
    def load_dataframe(self,dataframe):
        kwargs = self.__guess_headers(dataframe)
        print(kwargs)
        self.__load_dataframe(dataframe, **kwargs)
        
            
    def __load_dataframe(self, dataframe, cell_header = 'Cell_No', spot_header = 'Spot_No',time_header = 'Time_sec', red_header='Red_Int', green_header = 'Green_int', blue_header='Blue_int' ):
            df = dataframe
            
            self.head = df.head()
            n_cells = max(df[cell_header])+1
            max_spots =  max(df[spot_header])+1
            total_n_spots = 0
            data_struct = []
            

            for i in range(0,n_cells):
                i_g = []
                i_r = []
                i_b = []
                intensitys = []
                times = []
                for j in range(0,max_spots):
                    temp_len_spot =  df[(df[cell_header] == i) & (df[spot_header] ==j) ]
    
                    if len(temp_len_spot) >0:
                        time = df[(df[cell_header] ==i) & (df[spot_header] ==j)][time_header].values
                        intensity_red = df[(df[cell_header] ==i) & (df[spot_header] ==j)][red_header].values
                        intensity_green = df[(df[cell_header] ==i) & (df[spot_header] ==j)][green_header].values
                        intensity_blue = df[(df[cell_header] ==i) & (df[spot_header] ==j)][blue_header].values
                        i_g.append(intensity_green)
                        i_r.append(intensity_red)
                        i_b.append(intensity_blue)
                        
                        intensity = np.vstack((intensity_red,intensity_green,intensity_blue))
                        intensitys.append(intensity)
                        times.append(time)
                data_struct.append([i_r,i_g,i_b,intensitys,times ])
         
            time_arrays = []
            intensity_arrays = []
            cells = []
            for n in range(len(data_struct)):
                cells = cells + [n for x in range( len(data_struct[n][4])  )]
                time_arrays = time_arrays + data_struct[n][4]

            len_t = len(time_arrays[0])
            self.ragged = False
            for n in range(len(time_arrays)):
                if len(time_arrays[n]) != len_t:
                    self.ragged = True
                    
            self.intensity_vec = intensitys
            self.data_struct = data_struct
            self.times = time_arrays
            self.cells = cells               


class OptChain():
    '''
    Container class for parameter optimization chains
    '''
    def __init__(self):
        self.parchain = np.array([])
        self.iterations = 0
        self.parnames = np.array([])
        self.evalchain  = np.array([])
        self.objfunchain = np.array([])
        self.bestpar = None
        self.besteval = None
        self.opt_method = None
        self.opt_args = None
        self.objective_args = None
        self.objective_fun_list  = None
        self.intensity_fun = None
        self.logspace = False
        
    def report(self):
        print('=====================')
        print('Optimizer: %s ran for %d iterations ' % (self.opt_method,self.iterations))
        print('Optimizer arguments: ' + str(self.opt_args))
        print('Objective function: ' + str(self.objective_fun_list))
        print('Objective arguments: ' + str(self.objective_args))
        print('_____________________')
        print('Best Parameter Set: %s, feval: %d'%  (''.join(str(self.bestpar.tolist())),self.besteval ) )
        print('=====================')
        
    def check_parameter_convergence(self):
        
        def get_acc2(data, trunc=False):
            '''
            Get autocorrelation function
    
            *NOT* multi-tau
            '''
            N = len(data)
            fvi = np.fft.fft(data, n=2*N)
            acf = fvi*np.conjugate(fvi)
            acf = np.fft.ifft(acf)
            acf = np.real(acf[:N])/float(N)
            if trunc:
                acf[acf < 0]=0
                for i in range(1, len(acf)):
                    if acf[i] > acf[i-1]:
                        acf[i] = acf[i-1]
            return acf    
        
        for i in range(0,self.parchain.shape[1] ):
            acc = get_acc2(self.parchain[:,i] )       
            if i == 0:
                self.par_acc = acc
            else:
                self.par_acc = np.vstack( (self.par_acc, acc))
        self.par_acc = self.par_acc.T
            
    def check_objfun_convergence(self):
        
        def get_acc2(data, trunc=False):
            '''
            Get autocorrelation function
    
            *NOT* multi-tau
            '''
            N = len(data)
            fvi = np.fft.fft(data, n=2*N)
            acf = fvi*np.conjugate(fvi)
            acf = np.fft.ifft(acf)
            acf = np.real(acf[:N])/float(N)
            if trunc:
                acf[acf < 0]=0
                for i in range(1, len(acf)):
                    if acf[i] > acf[i-1]:
                        acf[i] = acf[i-1]
            return acf    
        
        for i in range(0,self.objfunchain.shape[1] ):
            acc = get_acc2(self.objfunchain[:,i] )       
            if i == 0:
                self.objfun_acc = acc
            else:
                self.objfun_acc = np.vstack( (self.objfun_acc, acc))
                
        self.objfun_acc = self.objfun_acc.T
        
    
    def __clear_invalid_values(self):
        evalchain = self.evalchain
        parchain = self.parchain
        trimmed_evals = evalchain[~np.isnan(evalchain)]        
        final_evals = trimmed_evals[np.isfinite(trimmed_evals)]
        trimmed_parchains = parchain[~np.isnan(evalchain)]
        final_parchains = trimmed_parchains[np.isfinite(trimmed_evals)]
        
        
        return final_evals, final_parchains
        
        
    def parplot(self,ellipse=True,logspace=False):
        n_par = len(self.bestpar)
       
        fig, ax = plt.subplots( n_par,n_par,dpi=300)
        
        plotnum = np.arange(n_par**2).reshape((n_par,n_par)) + 1
        triangle_inds = np.tril(plotnum)[np.where(np.tril(plotnum) !=0 )]
        nplots = len(triangle_inds)
        
        
        if np.sum(np.isnan(self.evalchain))>0 or np.sum(~np.isfinite(self.evalchain)) >0 :
            print('Warning: NaN or Infinite values detected within function evaluation chain, these parameter sets are left out of the plot ')
        eval_chain,par_chain = self.__clear_invalid_values()
         
        
        viridis = cm.get_cmap('viridis', int(np.ceil(np.max(eval_chain))))
        colors = eval_chain
        covariances = np.cov(par_chain.T)
        used_pairs = []
        for i in range(n_par-1,-1,-1):
            for j in range(n_par-1,-1,-1):
                if set([i,j]) not in used_pairs:
                    if i != j:
                        a = ax[i][j].scatter(par_chain[:,i],par_chain[:,j], marker='.',c= colors)
                        if ellipse == True:
                            self.__get_ellipse([np.mean(par_chain[:,i]),np.mean(par_chain[:,j])], covariances,ax=ax[i][j] )
                                         
                    else:
                        b = ax[i][j].hist(par_chain[:,i],bins=40,density=True)
                        
           
                    used_pairs.append(set([i,j]))
                else: 
                    ax[i][j].axis('off')
                
                if len(self.parnames) >0:
                    if i == n_par-1:
                        ax[i][j].set_xlabel(self.parnames[i])
                    if j == 0:
                        ax[i][j].set_ylabel(self.parnames[j])
                        
                if logspace:
                    if i != j:
                        ax[i][j].set_yscale('log')
                    ax[i][j].set_xscale('log')
 
        fig.tight_layout()
        fig.colorbar(a, ax=ax)
        fig.show()
        
    def __get_ellipse(self,mu,cov,ax=None,crosshairs=False,*args,**kwargs):
        ''' Command to plot the fit instance 
        '''
        cmap= cm.viridis
        cmap = cm.coolwarm
        ci = .95
        vals,vecs = np.linalg.eig(cov)
        theta = float(abs((360/(2*np.pi))*np.arctan(vecs[1,0]/vecs[0,0])))
        # able to change CI now.
        scale = chi2.ppf(ci,2)
        w = np.sqrt(vals[0]*scale)*2
        h = np.sqrt(vals[1]*scale)*2
        

        e = Ellipse(xy = tuple(mu) ,width = w, height = h,angle=90-theta,**kwargs )

        ax.add_artist(e)
    
        e.set_clip_box(ax.bbox)
        # e.set_alpha(.75)
        e.set_edgecolor(('red'))
        e.set_linestyle(('-'))
        e.set_facecolor(('none'))
        return ax
    
    def save(self,filename):
        ext = filename.split('.')[-1]
        if ext == 'txt':
            x=1
        if ext in ['p','pickle']:
            pickle.dump(self,open('filename','wb'))
        if ext == 'csv':
            x=1
        if ext == 'npz':
            x=1
            
    def load(self,filename):
        ext = filename.split('.')[-1]
        if ext == 'txt':
            x=1
        if ext in ['p','pickle']:
            tmp_obj = pickle.load(open('filename','wb'))
            for item in tmp_obj.__dict__.keys():
                self.__dict__[item] = tmp_obj.__dict__[item]
            
        if ext == 'csv':
            x=1
        if ext == 'npz':
            x=1
                    

class suite():
    '''
    a coterie of constructs
    '''

    def __init__(self):
        self.pois = []
        self.discernable = False
        self.models = []
        self.combo_list = []
        